# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
DumpDataParser class. This class mainly involves the parser_dump_data function.
"""
import os
import json
import struct

import numpy as np

from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.file_utils import FileUtils
from msprobe.msaccucmp.cmp_utils.multi_process.multi_convert_process import MultiConvertProcess
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse import dump_utils
from msprobe.msaccucmp.dump_parse.dump_data_object import DumpDataObj


class DumpDataParser:
    """
    The class for dump data parser
    """
    OLD_OVERFLOW_ELEMENT = ('model_id', 'stream_id', 'task_id', 'task_type', 'pc_start', 'para_base')

    def __init__(self: any, arguments: any) -> None:
        self.path_str = arguments.dump_path
        self.input_path = []
        self.multi_process = None
        if os.path.islink(os.path.abspath(arguments.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % arguments.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(arguments.output_path)
        self.dump_version = arguments.dump_version
        self.output_file_type = arguments.output_file_type

    @staticmethod
    def _parser_overflow_info(data: any, start: int) -> dict:
        overflow_info = {}
        for idx, key in enumerate(DumpDataParser.OLD_OVERFLOW_ELEMENT):
            index = start + idx * ConstManager.UINT64_SIZE
            value = OpDebugInfoParser.unpack_uint_value(data, index, 'UINT64')
            if key in ['pc_start', 'para_base']:
                overflow_info[key] = hex(value)
            else:
                overflow_info[key] = value
        return overflow_info

    @staticmethod
    def _parser_ai_core_status(ai_core_info: dict, data: any, start: int) -> None:
        index = start
        kernel_code = OpDebugInfoParser.unpack_uint_value(data, index, 'UINT64')
        ai_core_info['kernel_code'] = hex(kernel_code)
        index += ConstManager.UINT64_SIZE
        block_idx = OpDebugInfoParser.unpack_uint_value(data, index, 'UINT64')
        ai_core_info['block_idx'] = block_idx
        index += ConstManager.UINT64_SIZE
        status = OpDebugInfoParser.unpack_uint_value(data, index, 'UINT64')
        ai_core_info['status'] = status

    @staticmethod
    def _write_log(log_file_path, log_space):
        log_str = log_space.data.decode()
        log_str = log_str.replace(ConstManager.END_FLAG, "")
        path_check.check_write_path_secure(log_file_path)
        with os.fdopen(os.open(log_file_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES), 'w+') as text_file:
            text_file.truncate()
            text_file.write(log_str)

    def parse_dump_data(self: any) -> int:
        """
        Convert dump data to numpy and bin file
        """
        # 1. check arguments valid
        self.check_arguments_valid()
        # 2. parse dump data
        if len(self.input_path) == 1 and os.path.isfile(self.input_path[0]):
            ret, _ = self._parse_one_dump_file(self.input_path[0])
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                log.print_error_log('Failed to parse dump file "%s".' % self.input_path[0])
            return ret
        return self.multi_process.process()

    def parse_log_data(self: any) -> int:
        """
        Convert dump data to numpy and bin file
        """
        # 1. check arguments valid
        ret = path_check.check_output_path_valid(self.output_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        ret = path_check.check_path_valid(self.path_str, True, False, path_type=path_check.PathType.File)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        try:
            self._save_log_data()
        except CompareError as error:
            return error.code
        return CompareError.MSACCUCMP_NONE_ERROR

    def check_arguments_valid(self: any) -> None:
        """
        Check arguments valid
        """
        self.input_path = path_check.get_path_list_for_str(self.path_str)
        self.multi_process = MultiConvertProcess(self._parse_one_dump_file, self.input_path, self.output_path)
        ret = path_check.check_output_path_valid(self.output_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)

    def _save_log_data(self):
        dump_data = dump_utils.parse_dump_file(self.path_str, self.dump_version)
        log_space = dump_data.space
        dump_file_name = os.path.basename(self.path_str)
        if len(log_space) == 0:
            log.print_error_log("There is no log data in %r" % self.path_str)
            raise CompareError(CompareError.MSACCUCMP_INVALID_OVERFLOW_TYPE_ERROR)
        for (index, log_data) in enumerate(log_space):
            log_file_name = '%s.%d.log' % (dump_file_name, index)
            log_file_path = os.path.join(self.output_path, log_file_name)
            log.print_info_log('Start to parse the data of log:%d in "%r".' % (index, self.path_str))
            try:
                self._write_log(log_file_path, log_data)
            except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
                log.print_error_log('Failed to save log data. %s'
                                    % str(error))
                raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR) from error

            log.print_info_log('The data of log:%d has been parsed into "%r".'
                               % (index, log_file_path))

    def _save_tensor_to_file(self: any, dump_path: str, dump_data: DumpDataObj, tensor_type: str) -> None:
        tensor_list = dump_data.input_data if tensor_type == ConstManager.INPUT else dump_data.output_data
        op_name = dump_data.op_name
        ffts_auto = False
        if len(tensor_list) == 0:
            log.print_warn_log('There is no %s in "%r".' % (tensor_type, dump_path))
            return
        if dump_data.get_ffts_mode:
            if tensor_type == ConstManager.INPUT:
                shape_list = dump_data.ffts_auto_input_shape_list
            else:
                shape_list = dump_data.ffts_auto_output_shape_list
            if len(shape_list) == len(tensor_list):
                ffts_auto = True
            else:
                log.print_error_log('The length of shape_list %s is not equal \
                    to the length of tensor_list %s.' % (shape_list, tensor_list))
                raise CompareError(CompareError.MSACCUCMP_INVALID_INPUT_MAPPING)
                
        name = os.path.basename(dump_path)
        for (index, tensor) in enumerate(tensor_list):
            log.print_info_log('Start to parse the data of %s:%d in "%r".' % (tensor_type, index, dump_path))
            try:
                array = tensor.data
            except CompareError:
                log.print_warn_log('Failed to parse the data of %s:%d in "%r".' % (tensor_type, index, dump_path))
                continue
            if self.output_file_type == 'npy':
                file_name = '%s.%s.%d.npy' % (name, tensor_type, index)
                file_name = FileUtils.handle_too_long_file_name(
                    file_name, '.npy', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
            elif self.output_file_type == 'msnpy':
                file_name = utils.make_msnpy_file_name(dump_path, op_name, tensor_type, index, tensor.tensor_format)
                file_name = FileUtils.handle_too_long_file_name(
                    file_name, '.npy', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
            else:
                file_name = '%s.%s.%d.%s.%s.%s.bin' \
                            % (name, tensor_type, index, utils.get_string_from_list(array.shape, '_'),
                               np.dtype(common.get_dtype_by_data_type(tensor.data_type)).name,
                               common.get_format_string(tensor.tensor_format))
                file_name = FileUtils.handle_too_long_file_name(
                    file_name, '.bin', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
            output_file_path = os.path.join(self.output_path, file_name)
            if ffts_auto:
                FileUtils.save_array_to_file(
                    output_file_path, array, self.output_file_type != 'bin', shape_list[index])
            else:
                FileUtils.save_array_to_file(output_file_path, array, self.output_file_type != 'bin', tensor.shape)
            log.print_info_log('The data of %s:%d has been parsed into "%r".'
                               % (tensor_type, index, output_file_path))

    def _save_buffer_to_file(self: any, dump_path: str, tensor_list: list) -> None:
        if tensor_list is None or len(tensor_list) == 0:
            log.print_warn_log('There is no buffer data in "%r".' % dump_path)
            return
        name = os.path.basename(dump_path)
        for (index, tensor) in enumerate(tensor_list):
            buffer_type = ConstManager.BUFFER_TYPE_MAP.get(tensor.buffer_type)
            log.print_info_log('Start to parse the data of %sbuffer:%d in "%r".'
                               % (buffer_type, index, dump_path))
            file_name = "%s.%sbuffer.%s.bin" % (name, buffer_type, index)
            file_name = FileUtils.handle_too_long_file_name(
                file_name, '.bin', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
            output_dump_path = os.path.join(self.output_path, file_name)
            FileUtils.save_data_to_file(output_dump_path, tensor.data, 'wb', delete=True)
            log.print_info_log('The data of %sbuffer:%d has been parsed into "%r".'
                               % (buffer_type, index, output_dump_path))

    def _save_space_to_file(self: any, dump_path: str, tensor_list: list) -> None:
        if tensor_list is None or len(tensor_list) == 0:
            log.print_warn_log('There is no space data in "%s".' % dump_path)
            return
        name = os.path.basename(dump_path)
        for (index, tensor) in enumerate(tensor_list):
            log.print_info_log('Start to parse the data of space:%d in "%r".' % (index, dump_path))
            file_name = "%s.space.%s.bin" % (name, index)
            file_name = FileUtils.handle_too_long_file_name(
                file_name, '.bin', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
            output_dump_path = os.path.join(self.output_path, file_name)
            FileUtils.save_data_to_file(output_dump_path, tensor.data, 'wb', delete=True)
            log.print_info_log('The data of space:%d has been parsed into "%r".' % (index, output_dump_path))

    def _save_op_debug_to_file(self: any, dump_path: str, output: any) -> None:
        for idx, item in enumerate(output):
            bytes_data = utils.convert_ndarray_to_bytes(item.data)
            magic = OpDebugInfoParser.unpack_uint_value(bytes_data, 0, 'UINT32')
            if magic == ConstManager.MAGIC_NUM:
                debug_info_parser = OpDebugInfoParser(bytes_data)
                debug_info = debug_info_parser.parse_op_debug_new_version()
            else:
                debug_info = self._parse_op_debug_old_version(dump_path, idx, bytes_data)
            json_path = os.path.join(self.output_path, "%s.output.%d.json" % (os.path.basename(dump_path), idx))
            FileUtils.save_data_to_file(json_path, json.dumps(debug_info, sort_keys=False, indent=4), 'w+', delete=True)
            log.print_info_log('The data of output:%d has been parsed into "%r".' % (idx, json_path))

    def _parse_op_debug_old_version(self: any, dump_path: str, idx: int, item_data: any) -> dict:
        if len(item_data) != ConstManager.OVERFLOW_CHECK_SIZE:
            log.print_error_log('The data size (%d) of output:%d is not equal to %d in %r. '
                                'Please check the dump file.'
                                % (len(item_data), idx, ConstManager.OVERFLOW_CHECK_SIZE, dump_path))
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)
        # parser DHA Atomic Add info
        index = 0
        dha_atomic_add_info = self._parser_overflow_info(item_data, index)
        # parser L2 Atomic Add info
        index += ConstManager.DHA_ATOMIC_ADD_INFO_SIZE
        l2_atomic_add_info = self._parser_overflow_info(item_data, index)
        # parser AI Core info
        index += ConstManager.L2_ATOMIC_ADD_INFO_SIZE
        ai_core_info = self._parser_overflow_info(item_data, index)
        # parser DHA Atomic Add status
        index += ConstManager.AI_CORE_INFO_SIZE
        dha_atomic_add_status = OpDebugInfoParser.unpack_uint_value(item_data, index, 'UINT64')
        dha_atomic_add_info['status'] = dha_atomic_add_status
        # parser L2 Atomic Add status
        index += ConstManager.DHA_ATOMIC_ADD_STATUS_SIZE
        l2_atomic_add_status = OpDebugInfoParser.unpack_uint_value(item_data, index, 'UINT64')
        l2_atomic_add_info['status'] = l2_atomic_add_status
        # parser AI Core status
        index += ConstManager.L2_ATOMIC_ADD_STATUS_SIZE
        self._parser_ai_core_status(ai_core_info, item_data, index)

        data = {
            'DHA Atomic Add': dha_atomic_add_info,
            'L2 Atomic Add': l2_atomic_add_info,
            'AI Core': ai_core_info
        }
        return data

    def _parse_one_file_exec(self: any, dump_path: str) -> None:
        ret = path_check.check_path_valid(dump_path, True, False, path_type=path_check.PathType.File)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        dump_data = dump_utils.parse_dump_file(dump_path, self.dump_version)
        
        if os.path.basename(dump_path).startswith('Opdebug.Node_OpDebug.'):
            self._save_op_debug_to_file(dump_path, dump_data.output_data)
        else:
            if dump_data.get_ffts_mode:
                try:
                    thread_id = int(dump_path.split('.')[-2])
                except (IndexError, ValueError) as e:
                    log.print_error_log('Parse thread_id failed, please check dump_path! dump_path: {}'
                                        .format(dump_path))
                    raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR) from e
                dump_data.ffts_auto_input_shape_list = dump_data.calculate_auto_mode_shape(thread_id, "input")    
                dump_data.ffts_auto_output_shape_list = dump_data.calculate_auto_mode_shape(thread_id, "output")
            
            self._save_tensor_to_file(dump_path, dump_data, 'input')
            self._save_tensor_to_file(dump_path, dump_data, 'output')
            self._save_buffer_to_file(dump_path, dump_data.buffer)
            self._save_space_to_file(dump_path, dump_data.space)

    def _parse_one_dump_file(self: any, dump_path: str) -> (int, str):
        try:
            self._parse_one_file_exec(dump_path)
        except CompareError as error:
            return error.code, dump_path
        return CompareError.MSACCUCMP_NONE_ERROR, dump_path


class OpDebugInfoParser:
    """
    Data parsing class for OpDebug dump files of the new version
    """

    def __init__(self: any, data: any) -> None:
        self.data = data
        self.overflow_info = {}

    @staticmethod
    def unpack_uint_value(data: any, index: int, uint_type: str) -> int:
        type_para = ConstManager.UNPACK_FORMAT.get(uint_type)
        # 返回值是数据类型名字,按照数据字长取地址得到的数据
        return struct.unpack(type_para.get('FMT'), data[index:index + type_para.get('SIZE')])[0]

    @staticmethod
    def _change_debug_info_fomat(debug_info: dict) -> None:
        for item in ConstManager.HEX_FORMAT_ITEM:
            if item in debug_info.keys():
                debug_info[item] = hex(debug_info[item])

    @staticmethod
    def is_new_type(acc_type: str, version: int):
        return acc_type in ("AIC", "AIV") and version == 1

    def _check_acc_debug_info(self, acc_debug_info: dict, version: int) -> None:
        if acc_debug_info.get('valid') != 1:
            log.print_error_log('The value of valid in the OpDebug file is {}, it is not 1'
                                .format(acc_debug_info.get('valid')))

        acc_type = acc_debug_info.get('acc_type')
        if not acc_type:
            log.print_error_log("The value of 'acc_type' is {}, it's invalid".format(acc_type))
            raise CompareError(CompareError.MSACCUCMP_INVALID_OVERFLOW_TYPE_ERROR)

        expect_data_len = 0
        if acc_type in ConstManager.DEBUG_INFO_MAP.keys():
            if self.is_new_type(acc_type, version):
                expect_data_len = len(ConstManager.DEBUG_INFO_MAP.get(acc_type)) + 3
            else:
                expect_data_len = len(ConstManager.DEBUG_INFO_MAP.get(acc_type))
        expect_data_len *= ConstManager.UINT64_SIZE

        real_data_len = acc_debug_info.get('data_len')
        if real_data_len != expect_data_len:
            log.print_error_log('The value of data_len in the OpDebug file is {}, it is not equal'
                                ' the expect value {}'.format(real_data_len, expect_data_len))

    def parse_op_debug_new_version(self: any) -> dict:
        op_debug = {}

        # 先解析所有 UINT32 字段（包括 version）
        for idx, key in enumerate(ConstManager.OVERFLOW_DEBUG):
            if key == 'acc_list':
                continue
            index = idx * ConstManager.UINT32_SIZE
            value = self.unpack_uint_value(self.data, index, 'UINT32')
            op_debug[key] = value
        version = op_debug.get('version', None)
        
        # 再解析 acc_list
        acc_index = ConstManager.OVERFLOW_DEBUG.index('acc_list') * ConstManager.UINT32_SIZE
        op_debug['acc_list'] = self._parse_acc_debug_info(acc_index, version)

        magic_key = ConstManager.MAGIC_KEY_WORD
        op_debug[magic_key] = hex(op_debug.get(magic_key))
        return op_debug

    def _parse_acc_debug_info(self: any, start: int, version: int) -> dict:
        acc_debug_info = {}
        acc_type = 0
        data_index = 0
        for idx, key in enumerate(ConstManager.ACC_DEBUG):
            index = start + idx * ConstManager.UINT32_SIZE
            if key == 'acc_type':
                type_value = self.unpack_uint_value(self.data, index, 'UINT32')
                acc_type = ConstManager.ACC_TYPE.get(type_value)
                acc_debug_info[key] = acc_type
                continue
            if key == 'data':
                data_index = index
            else:
                value = self.unpack_uint_value(self.data, index, 'UINT32')
                acc_debug_info[key] = value

        self._check_acc_debug_info(acc_debug_info, version)
        acc_debug_info['data'] = self._parse_debug_info(data_index, acc_type, version)
        return acc_debug_info

    def _parse_new_type_debug_info(self, data_names: tuple, start: int) -> dict:
        debug_info = {}
        for idx, key in enumerate(data_names):
            if key == 'status':
                debug_info[key] = self._parse_new_type_status_field(start, idx)
            else:
                index = start + idx * ConstManager.UINT64_SIZE
                debug_info[key] = self.unpack_uint_value(self.data, index, 'UINT64')
        return debug_info

    def _parse_new_type_status_field(self, start: int, idx: int) -> list:
        base_index = start + idx * ConstManager.UINT64_SIZE
        return [
            self.unpack_uint_value(self.data, base_index + i * ConstManager.UINT64_SIZE, 'UINT64')
            for i in range(ConstManager.STATUS_LEN)
        ]

    def _parse_normal_debug_info(self, data_names: tuple, start: int) -> dict:
        """解析普通类型字段"""
        return {
            key: self.unpack_uint_value(self.data, start + idx * ConstManager.UINT64_SIZE, 'UINT64')
            for idx, key in enumerate(data_names)
        }

    def _parse_debug_info(self: any, start: int, acc_type: str, version: int) -> dict:
        data_names = []
        if acc_type in ConstManager.DEBUG_INFO_MAP.keys():
            data_names = ConstManager.DEBUG_INFO_MAP.get(acc_type)

        debug_info = {}
        if self.is_new_type(acc_type, version):
            debug_info = self._parse_new_type_debug_info(data_names, start)
        else:
            debug_info = self._parse_normal_debug_info(data_names, start)
        self._change_debug_info_fomat(debug_info)
        return debug_info



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import concurrent
import copy
import os
import threading
import traceback
from datetime import datetime, timezone, timedelta

from msprobe.core.common.const import Const
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.file_utils import FileOpen, save_json, check_path_before_create
from msprobe.core.common.log import logger

lock = threading.Lock()


class DataWriter:

    def __init__(self) -> None:
        self.bench_dump_file_path = None
        self.dump_file_path = None
        self.stack_file_path = None
        self.construct_file_path = None
        self.dump_tensor_data_dir = None
        self.debug_file_path = None
        self.dump_error_info_path = None
        self.flush_size = 1000
        self.md5_flush_size = 5000
        self.larger_flush_size = 20000
        self.cache_data = {}
        self.cache_stack = {}
        self.cache_construct = {}
        self.cache_debug = {}
        self.stat_stack_list = []
        self._error_log_initialized = False
        self._cache_logged_error_types = set()
        self.crc32_stack_list = []
        self.data_updated = False

    @recursion_depth_decorator("JsonWriter: DataWriter._replace_crc32_placeholders")
    def _replace_crc32_placeholders(self, data, crc32_results):
        """
        遍历 JSON 结构，将所有 md5_index 占位符替换成真实的 CRC32
        """
        if isinstance(data, dict):
            for k, v in list(data.items()):
                if k == Const.MD5_INDEX and isinstance(v, int):
                    idx = v
                    # 防越界
                    crc = crc32_results[idx] if idx < len(crc32_results) else None
                    # 删除占位符，改成真实字段
                    del data[k]
                    data[Const.MD5] = crc
                else:
                    self._replace_crc32_placeholders(v, crc32_results)
        elif isinstance(data, (list, tuple)):
            for item in data:
                self._replace_crc32_placeholders(item, crc32_results)

    @recursion_depth_decorator("JsonWriter: DataWriter._replace_stat_placeholders")
    def _replace_stat_placeholders(self, data, stat_result):
        if isinstance(data, dict):
            keys = list(data.keys())  # 获取当前所有键
            for key in keys:  # 递归所有变量
                value = data[key]
                if key == Const.TENSOR_STAT_INDEX and isinstance(value, int):
                    if value >= 0:
                        idx = value
                    else:
                        return
                    stat_values = stat_result[idx] if idx < len(stat_result) else [None] * 4

                    new_entries = {
                        Const.TYPE: data["type"],
                        Const.DTYPE: data["dtype"],
                        Const.SHAPE: data["shape"],
                        Const.MAX: stat_values[0],
                        Const.MIN: stat_values[1],
                        Const.MEAN: stat_values[2],
                        Const.NORM: stat_values[3],
                    }
                    if Const.MS_FRAMEWORK in data["type"]:
                        layout = data.get("layout", None)
                        if layout:
                            new_entries[Const.LAYOUT] = layout
                        shard_size = data.get("hsdp_shard_size", None)
                        if shard_size:
                            new_entries[Const.SHARD_SIZE] = shard_size
                    del data[key]

                    # 重构字典顺序
                    updated_dict = {}
                    # 通过插入排序后字段保证字段写入json的有序
                    updated_dict.update(new_entries)
                    # 遍历原字典其他字段（排除已删除的tensor_stat_index）
                    for k in data:
                        if k not in new_entries:
                            updated_dict[k] = data[k]
                    data.clear()
                    data.update(updated_dict)
                else:
                    self._replace_stat_placeholders(value, stat_result)
        elif isinstance(data, (list, tuple)):
            for item in data:
                self._replace_stat_placeholders(item, stat_result)

    def reset_cache(self):
        self.cache_data = {}
        self.cache_stack = {}
        self.cache_construct = {}
        self.cache_debug = {}
        self._cache_logged_error_types = set()

    def append_crc32_to_buffer(self, future: concurrent.futures.Future) -> int:
        """
        把一个计算 CRC32 的 Future 放入队列，返回占位符索引
        """
        idx = len(self.crc32_stack_list)
        self.crc32_stack_list.append(future)
        return idx

    def flush_crc32_stack(self):
        """
        等待所有 CRC32 计算完成，返回结果列表
        """
        if not self.crc32_stack_list:
            return []
        results = [f.result() for f in self.crc32_stack_list]
        self.crc32_stack_list = []
        return results

    def initialize_json_file(self, **kwargs):
        if kwargs["level"] == Const.LEVEL_DEBUG and not self.cache_debug:
            # debug level case only create debug.json
            debug_dict = copy.deepcopy(kwargs)
            debug_dict.update({"dump_data_dir": self.dump_tensor_data_dir, Const.DATA: {}})
            self.cache_debug = debug_dict
            save_json(self.debug_file_path, self.cache_debug, indent=1)
            return
        if not self.cache_data:
            kwargs.update({"dump_data_dir": self.dump_tensor_data_dir, Const.DATA: {}})
            self.cache_data = kwargs
            save_json(self.dump_file_path, self.cache_data, indent=1)
        if not self.cache_stack:
            save_json(self.stack_file_path, self.cache_stack, indent=1)
        if not self.cache_construct:
            save_json(self.construct_file_path, self.cache_construct, indent=1)

    def update_dump_paths(self, dump_path_aggregation):
        self.dump_file_path = dump_path_aggregation.dump_file_path
        self.stack_file_path = dump_path_aggregation.stack_file_path
        self.construct_file_path = dump_path_aggregation.construct_file_path
        self.dump_tensor_data_dir = dump_path_aggregation.dump_tensor_data_dir
        self.debug_file_path = dump_path_aggregation.debug_file_path
        self.bench_dump_file_path = getattr(dump_path_aggregation, "bench_dump_file_path", None)
        self.dump_error_info_path = dump_path_aggregation.dump_error_info_path


    def flush_data_periodically(self):
        dump_data = self.cache_data.get(Const.DATA)

        if not dump_data or not isinstance(dump_data, dict):
            return

        length = len(dump_data)

        # 1) 先取到 config（如果没有，就拿 None）
        cfg = getattr(self, "config", None)
        # 2) 再取 summary_mode（如果 cfg 是 None 或者没 summary_mode，就拿 None）
        summary_mode = getattr(cfg, "summary_mode", None)

        if summary_mode == Const.MD5:
            threshold = self.md5_flush_size
        else:
            threshold = self.flush_size if length < self.larger_flush_size else self.larger_flush_size

        if length % threshold == 0:
            self.write_json()

    def write_error_log(self, message: str, error_type: str):
        """
        写错误日志：
          - 第一次调用时以 'w' 模式清空文件，之后都用 'a' 模式追加
          - 添加时间戳
          - 在 message 后写入当前的调用栈（方便追踪日志来源）
        """
        # 如果同类型错误已经记录过，跳过
        if error_type in self._cache_logged_error_types:
            return
        # 否则添加到已记录集合，并继续写日志
        self._cache_logged_error_types.add(error_type)

        try:
            mode = "w" if not self._error_log_initialized else "a"
            self._error_log_initialized = True

            check_path_before_create(self.dump_error_info_path)

            with FileOpen(self.dump_error_info_path, mode) as f:
                cst_timezone = timezone(timedelta(hours=8), name="CST")
                timestamp = datetime.now(cst_timezone).strftime("%Y-%m-%d %H:%M:%S %z")
                f.write(f"[{timestamp}] {message}\n")
                f.write("Call stack (most recent call last):\n")

                f.write("".join(traceback.format_stack()[:-1]))  # 去掉自己这一层
                f.write("\n")
        except Exception as e:
            # 如果连写日志都失败了，就打印到 stderr
            logger.warning(f"[FallbackError] Failed to write error log: {e}")

    def update_data(self, new_data):
        with lock:
            if not isinstance(new_data, dict) or len(new_data.keys()) != 1:
                logger.warning(f"The data info({new_data}) should be a dict with only one outer key.")
                return
            dump_data = self.cache_data.get(Const.DATA)
            if not isinstance(dump_data, dict):
                logger.warning(f"The dump data({dump_data}) should be a dict.")
                return

            self.data_updated = True
            key = next(iter(new_data.keys()))
            if key in dump_data:
                dump_data.get(key).update(new_data.get(key))
            else:
                dump_data.update(new_data)

    def update_stack(self, name, stack_data):
        with lock:
            self.data_updated = True
            api_list = self.cache_stack.get(stack_data)
            if api_list is None:
                self.cache_stack.update({stack_data: [name]})
            else:
                api_list.append(name)

    def update_construct(self, new_data):
        with lock:
            self.data_updated = True
            self.cache_construct.update(new_data)

    def update_debug(self, new_data):
        with lock:
            self.data_updated = True
            self.cache_debug['data'].update(new_data)

    def write_data_json(self, file_path):
        logger.info(f"dump.json is at {os.path.dirname(os.path.dirname(file_path))}. ")
        save_json(file_path, self.cache_data, indent=1)

    def write_stack_info_json(self, file_path):
        num, new_cache_stack = 0, {}
        for key, value in self.cache_stack.items():
            new_cache_stack[num] = [value, key]
            num += 1
        save_json(file_path, new_cache_stack, indent=1)

    def write_construct_info_json(self, file_path):
        save_json(file_path, self.cache_construct, indent=1)

    def write_debug_info_json(self, file_path):
        save_json(file_path, self.cache_debug, indent=1)

    def append_stat_to_buffer(self, stat_vector):
        """
        直接使用 Python list 存储 stat_vector,
        将 stat_vector 存入 self.stat_stack_list 的方式
        """
        self.stat_stack_list.append(stat_vector)
        return len(self.stat_stack_list) - 1

    def get_buffer_values_max(self, index):
        if 0 <= index < len(self.stat_stack_list) and len(self.stat_stack_list[index]) >= 1:
            return self.stat_stack_list[index][0]
        else:
            logger.warning(f"stat_stack_list[{index}] The internal data is incomplete,"
                           f" and the maximum value cannot be obtained.")
            return None

    def get_buffer_values_min(self, index):
        if 0 <= index < len(self.stat_stack_list) and len(self.stat_stack_list[index]) >= 1:
            return self.stat_stack_list[index][1]
        else:
            logger.warning(f"stat_stack_list[{index}] Internal data is incomplete"
                           f" and minimum values cannot be obtained.")
            return None

    def flush_stat_stack(self):
        """
        在 flush 阶段，将所有存储的统计值从设备搬到 CPU，
        这里返回一个列表，每个元素是 [Max, Min, Mean, Norm] 的数值列表
        """
        if not self.stat_stack_list:
            return []
        result = [
            [
                x.item() if hasattr(x, "item") else x
                for x in stat_values
            ]
            for stat_values in self.stat_stack_list
        ]
        self.stat_stack_list = []
        return result

    def write_json(self):
        with lock:
            # 在写 JSON 前，统一获取统计值
            stat_result = self.flush_stat_stack()
            # 遍历 cache_data，将占位符替换为最终统计值
            if stat_result:
                self.data_updated = True
                self._replace_stat_placeholders(self.cache_data, stat_result)
                if self.cache_debug:
                    self._replace_stat_placeholders(self.cache_debug, stat_result)

            crc32_result = self.flush_crc32_stack()
            if crc32_result:
                self.data_updated = True
                self._replace_crc32_placeholders(self.cache_data, crc32_result)
                if self.cache_debug:
                    self._replace_crc32_placeholders(self.cache_debug, crc32_result)

            if not self.data_updated:
                return

            if self.cache_data:
                self.write_data_json(self.dump_file_path)
            if self.cache_stack:
                self.write_stack_info_json(self.stack_file_path)
            if self.cache_construct:
                self.write_construct_info_json(self.construct_file_path)
            if self.cache_debug:
                self.write_debug_info_json(self.debug_file_path)
            self.data_updated = False



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import os
import stat

import numpy as np


class Const:
    """
    Class for const
    """
    TOOL_NAME = "msprobe"
    MD5_INDEX = "md5_index"
    MD5 = "md5"

    ipv4_pattern = "([1-9]?\d|1\d{2}|2[0-4]\d|25[0-5])(\.([1-9]?\d|1\d{2}|2[0-4]\d|25[0-5])){3}$"
    SEP = "."
    REGEX_PREFIX_MAX_LENGTH = 20
    REGEX_PREFIX_PATTERN = r"^[a-zA-Z0-9_-]+$"
    REGEX_FORWARD_BACKWARD = r'\.(forward|backward)\.'
    FILE_PATTERN = r'^[a-zA-Z0-9_./-]+$'
    STRING_BLACKLIST = r"^[＋－＝％＠\+\-=%@]|;[＋－＝％＠\+\-=%@]"
    COMMA = ","
    FLOAT_EPSILON = np.finfo(float).eps
    OFF = 'OFF'
    BACKWARD = 'backward'
    FORWARD = 'forward'
    PROGRESS_TIMEOUT = 3000
    EXCEPTION_NONE = None
    JIT = 'Jit'
    PRIMITIVE_PREFIX = 'Primitive'
    DEFAULT_LIST = []
    DEFAULT_PATH = './'
    WHITE_LIST = 'white_list'
    BLACK_LIST = 'black_list'
    DUMP_TENSOR_DATA = 'dump_tensor_data'
    NONE = None
    THREE_SEGMENT = 3
    FOUR_SEGMENT = 4
    SIX_SEGMENT = 6
    SEVEN_SEGMENT = 7

    MAX_DEPTH = 400
    CPU_QUARTER = 4
    DUMP_MAX_DEPTH = 400

    EXTERN_INPUT_LIST_MAX_LEN = 100
    MAX_PROCESS_NUM = 128

    # dump mode
    ALL = "all"
    LIST = "list"
    RANGE = "range"
    STACK = "stack"
    ACL = "acl"
    API_LIST = "api_list"
    API_STACK = "api_stack"
    DUMP_MODE = [ALL, LIST, RANGE, STACK, ACL, API_LIST, API_STACK]
    AUTO = "auto"
    ONLINE_DUMP_MODE = [ALL, LIST, AUTO, OFF]
    SUMMARY = "summary"
    MD5 = "md5"
    HASH = "hash"
    VALUE = "value"
    SUMMARY_MODE = ["statistics", "md5"]

    WRITE_FLAGS = os.O_WRONLY | os.O_CREAT
    WRITE_MODES = stat.S_IWUSR | stat.S_IRUSR
    OVERWRITE_FLAGS = os.O_WRONLY | os.O_CREAT | os.O_TRUNC

    PKL_SUFFIX = ".pkl"
    NUMPY_SUFFIX = ".npy"
    NUMPY_PATTERN = "*.npy"
    PT_SUFFIX = ".pt"
    PY_SUFFIX = ".py"
    INIT_PY = "init.py"
    ONE_GB = 1073741824  # 1 * 1024 * 1024 * 1024
    TEN_GB = 10737418240  # 10 * 1024 * 1024 * 1024
    ONE_MB = 1048576  # 1 * 1024 * 1024
    FILE_PATTERN = r'^[a-zA-Z0-9_./-]+$'
    DISTRIBUTED_PREFIX_LENGTH = 60
    # env dump path
    KWARGS = 'kwargs'
    INPUT = 'input'
    OUTPUT = 'output'
    INPUT_ARGS = 'input_args'
    INPUT_KWARGS = 'input_kwargs'
    GRAD_INPUT = 'grad_input'
    GRAD_OUTPUT = 'grad_output'
    PARAMS = 'parameters'
    PARAMS_GRAD = 'parameters_grad'
    DEBUG = 'debug'
    START = "start"
    STOP = "stop"
    ENV_ENABLE = "1"
    ENV_DISABLE = "0"
    MAX_SEED_VALUE = 4294967295  # 2**32 - 1
    STATISTICS = "statistics"
    TENSOR = "tensor"
    OVERFLOW_CHECK = "overflow_check"
    DIFF_CHECK = "diff_check"
    FREE_BENCHMARK = "free_benchmark"
    ACC_CHECK = "acc_check"
    STRUCTURE = "structure"
    EXCEPTION_DUMP = "exception_dump"
    DUMP_PRECISION_HIGH = "high"
    DUMP_PRECISION_LOW = "low"
    TASK_LIST = [TENSOR, STATISTICS, ACC_CHECK, STRUCTURE, OVERFLOW_CHECK, DIFF_CHECK, EXCEPTION_DUMP]
    TORCH_TASK_LIST = [TENSOR, STATISTICS, ACC_CHECK, STRUCTURE, DIFF_CHECK]
    DUMP_DATA_COLLECTION_LIST = [STATISTICS, TENSOR, STRUCTURE]
    DUMP_DATA_MODE_LIST = [ALL, INPUT, OUTPUT, FORWARD, BACKWARD]
    DUMP_PRECISION_LIST = [DUMP_PRECISION_LOW, DUMP_PRECISION_HIGH]
    LEVEL_L0 = "L0"
    LEVEL_L1 = "L1"
    LEVEL_L2 = "L2"
    LEVEL_MIX = "mix"
    LEVEL_DEBUG = "debug"
    LEVEL_LIST = [LEVEL_L0, LEVEL_L1, LEVEL_L2, LEVEL_MIX, LEVEL_DEBUG]
    ATTR_NAME_PREFIX = "wrap_"
    ATTR_NAME_PREFIX_LEN = len(ATTR_NAME_PREFIX)
    KERNEL_DUMP = "kernel_dump"
    DATA = "data"
    PT_FRAMEWORK = "pytorch"
    MS_FRAMEWORK = "mindspore"
    MT_FRAMEWORK = "mindtorch"
    UNKNOWN_FRAMEWORK = "unknown"
    DIRECTORY_LENGTH = 4096
    FILE_NAME_LENGTH = 255
    FLOAT_TYPE = [np.half, np.single, float, np.double, np.float64, np.longdouble, np.float32, np.float16]
    BOOL_TYPE = [bool, np.uint8]
    INT_TYPE = [np.int32, np.int64]
    NPU = 'NPU'
    NPU_LOWERCASE = 'npu'
    CPU_LOWERCASE = 'cpu'
    GPU_LOWERCASE = 'gpu'
    CUDA_LOWERCASE = 'cuda'
    DEVICE = 'device'
    DISTRIBUTED = 'Distributed'
    DUMP_PREFIX = ["Distributed", "Functional", "Torch", "Tensor", "Mint", "MintFunctional", "Primitive",
                   "Aten", "VF", "NPU", "Jit"]
    MODULE_PREFIX = ["Module", "Cell"]
    FORWARD_NAME_SUFFIX = ".forward"

    DUMP_JSON_FILE = "dump_json_file"
    DEBUG_JSON_FILE = "debug_json_file"
    STACK_JSON_FILE = "stack_json_file"

    # struct json param
    ORIGIN_DATA = "origin_data"
    SCOPE = "scope"
    STACK = "stack"

    ATEN = "Aten"
    MODULE_WHITE_LIST = ["torch", "numpy"]

    FUNC_SKIP_LIST = ["construct", "__call__"]
    FILE_SKIP_LIST = ["msprobe", "MindSpeed"]
    DATA_TYPE_SKIP_LIST = ["Primitive", "Jit"]

    STACK_FILE_INDEX = 0
    STACK_FUNC_INDEX = 2
    STACK_FUNC_ELE_INDEX = 1

    SCOPE_ID_INDEX = -1
    SCOPE_DIRECTION_INDEX = -2
    TYPE_NAME_INDEX = -3
    PARAMS_GRAD_TYPE_NAME_INDEX = -2
    LAYER_NAME_INDEX = -4
    PARAMS_GRAD_NAME_INDEX = -3
    API_TYPE_INDEX = 0
    LEFT_MOVE_INDEX = -1
    RIGHT_MOVE_INDEX = 1
    LAST_INDEX = -1

    TOP_LAYER = "TopLayer"
    CELL = "Cell"
    MODULE = "Module"
    API = "api"
    PYNATIVE_MODE = "pynative"
    PYNATIVE_GRAPH_MODE = "pynative_graph"

    FRAME_FILE_LIST = ["site-packages/torch", "package/torch", "site-packages/mindspore", "package/mindspore"]
    INPLACE_LIST = [
        "broadcast", "all_reduce", "reduce", "all_gather", "gather", "scatter", "reduce_scatter",
        "_reduce_scatter_base", "_all_gather_base", "send", "recv", "irecv", "isend", "all_to_all_single", "all_to_all",
        "all_gather_into_tensor", "reduce_scatter_tensor"
    ]

    CONVERT = {
        "int32_to_int64": ["torch.int32", "torch.int64"],
        "int64_to_fp32": ["torch.int64", "torch.float32"]
    }

    CONVERT_API = {
        "int32_to_int64": ["cross_entropy"],
        "int64_to_fp32": ["histc"]
    }

    FA_SPECIAL_SPARSE_MODE = [2, 3, 4]

    FILL_CHAR_NUMS = 50
    TOOL_ENDS_SUCCESSFULLY = f"{TOOL_NAME} ends successfully."

    WITHOUT_CALL_STACK = "The call stack retrieval failed."
    STACK_FILTER_KEYWORDS = ["msprobe/core", "msprobe/pytorch", "msprobe/mindspore"]
    CALL_STACK_FLAG = "api_dump/api_registry"
    NEW_STACK_FLAG = "0"

    STEP = "step"
    RANK = "rank"
    HYPHEN = "-"
    STEP_RANK_MINIMUM_VALUE = 0
    STEP_RANK_MAXIMUM_VALUE = int(1e6)

    # data type const
    TORCH_INT_DTYPE = ["torch.int8", "torch.int32", "torch.int64"]
    TORCH_FLOAT_DTYPE = ["torch.bfloat16", "torch.float16", "torch.float32", "torch.float64"]
    FLOAT16 = "Float16"
    FLOAT32 = "Float32"
    BFLOAT16 = "BFloat16"
    TORCH_FLOAT16 = "torch.float16"
    TORCH_FLOAT32 = "torch.float32"
    TORCH_BFLOAT16 = "torch.bfloat16"

    TYPE = 'type'
    DTYPE = 'dtype'
    SHAPE = 'shape'
    LAYOUT = 'layout'
    SHARD_SIZE = 'hsdp_shard_size'
    STACK_INFO = 'stack_info'
    MAX = 'Max'
    MIN = 'Min'
    MEAN = 'Mean'
    NORM = 'Norm'
    DATA_NAME = 'data_name'
    STATE = 'state'
    REQ_GRAD = 'requires_grad'
    API_ORIGIN_NAME = 'api_origin_name'
    TENSOR_STAT_INDEX = 'tensor_stat_index'
    SUMMARY_METRICS_LIST = [MAX, MIN, MEAN, NORM]

    CODE_STACK = 'Code Stack'
    OP_NAME = 'Op Name'
    SCOPE_NAME = 'Scope Name'
    CODE_STACKS = 'Code Stacks'
    FILE_PATH = 'File Path'
    NEW_LINE = '\n'
    CSV_NEWLINE_SEPARATOR = ',\n'
    # 分隔符常量
    SCOPE_SEPARATOR = "/"
    REPLACEMENT_CHARACTER = "_"
    PIPE_SEPARATOR = "|"

    FORWARD_PATTERN = SEP + FORWARD + SEP
    BACKWARD_PATTERN = SEP + BACKWARD + SEP

    OPTIMIZER = "optimizer"
    CLIP_GRAD = "clip_grad"
    END_PREFIX = "end_"

    TENSOR_STAT_LEN = 2

    TENSOR_TYPE = "torch.Tensor"
    DTENSOR_TYPE = "torch.distributed.tensor.DTensor"
    FAKE_TENSOR_TYPE = "torch._subclasses.fake_tensor.FakeTensor"
    AC_TENSOR_TYPE = "torch.distributed._functional_collectives.AsyncCollectiveTensor"
    NESTED_TENSOR_TYPE = "torch.nested._internal.nested_tensor.NestedTensor"

    SUPPORT_API_FILE_NAME = "support_wrap_ops.yaml"

    API_ATTR_LIST = ["__name__", "default"]

    PT_API_TYPE_FUNCTIONAL = "functional"
    PT_API_TYPE_TENSOR = "tensor"
    PT_API_TYPE_TORCH = "torch"
    PT_API_TYPE_VF = "_VF"
    PT_API_TYPE_NPU = "torch_npu"
    PT_API_TYPE_ATEN = "aten"
    PT_API_TYPE_DIST = "distributed"
    PT_API_TYPE_DIST_C10D = "distributed_c10d"
    PT_API_TYPE_NPU_DIST = "npu_distributed"
    PT_API_TYPE_MINDSPEED = "mindspeed"

    MS_API_TYPE_OPS = "ops"
    MS_API_TYPE_TENSOR = "tensor"
    MS_API_TYPE_STUB_TENSOR = "stubtensor"
    MS_API_TYPE_MINT = "mint.ops"
    MS_API_TYPE_MINT_FUNC = "mint.nn.functional"
    MS_API_TYPE_COM = "communication.comm_func"
    MS_API_TYPE_MINT_DIST = "mint.distributed"

    FUNCTIONAL_API_TYPE_PREFIX = "Functional"
    TENSOR_API_TYPE_PREFIX = "Tensor"
    DIST_API_TYPE_PREFIX = "Distributed"

    TORCH_API_TYPE_PREFIX = "Torch"
    NPU_API_TYPE_PREFIX = "NPU"
    ATEN_API_TYPE_PREFIX = "Aten"
    VF_API_TYPE_PREFIX = "VF"
    MINDSPEED_API_TYPE_PREFIX = "MindSpeed"

    MINT_API_TYPE_PREFIX = "Mint"
    MINT_FUNC_API_TYPE_PREFIX = "MintFunctional"
    MINT_DIST_API_TYPE_PREFIX = "MintDistributed"

    SUPPORT_API_DICT_KEY_MAP = {
        PT_FRAMEWORK: {
            PT_API_TYPE_FUNCTIONAL: PT_API_TYPE_FUNCTIONAL,
            PT_API_TYPE_TENSOR: PT_API_TYPE_TENSOR,
            PT_API_TYPE_TORCH: PT_API_TYPE_TORCH,
            PT_API_TYPE_VF: PT_API_TYPE_VF,
            PT_API_TYPE_NPU: PT_API_TYPE_NPU,
            PT_API_TYPE_ATEN: PT_API_TYPE_ATEN,
            PT_API_TYPE_DIST: PT_API_TYPE_DIST,
            PT_API_TYPE_DIST_C10D: PT_API_TYPE_DIST,
            PT_API_TYPE_NPU_DIST: PT_API_TYPE_NPU_DIST,
            PT_API_TYPE_MINDSPEED: PT_API_TYPE_MINDSPEED
        },
        MS_FRAMEWORK: {
            MS_API_TYPE_OPS: MS_API_TYPE_OPS,
            MS_API_TYPE_TENSOR: MS_API_TYPE_TENSOR,
            MS_API_TYPE_STUB_TENSOR: MS_API_TYPE_TENSOR,
            MS_API_TYPE_MINT: MS_API_TYPE_MINT,
            MS_API_TYPE_MINT_FUNC: MS_API_TYPE_MINT_FUNC,
            MS_API_TYPE_COM: MS_API_TYPE_COM,
            MS_API_TYPE_MINT_DIST: MS_API_TYPE_MINT_DIST
        },
        MT_FRAMEWORK: {
            PT_API_TYPE_FUNCTIONAL: PT_API_TYPE_FUNCTIONAL,
            PT_API_TYPE_TENSOR: PT_API_TYPE_TENSOR,
            PT_API_TYPE_TORCH: PT_API_TYPE_TORCH,
            PT_API_TYPE_NPU: PT_API_TYPE_NPU,
            PT_API_TYPE_DIST: PT_API_TYPE_DIST
        }
    }

    API_DATA_PREFIX = {
        PT_FRAMEWORK: {
            PT_API_TYPE_FUNCTIONAL: FUNCTIONAL_API_TYPE_PREFIX,
            PT_API_TYPE_TENSOR: TENSOR_API_TYPE_PREFIX,
            PT_API_TYPE_TORCH: TORCH_API_TYPE_PREFIX,
            PT_API_TYPE_VF: VF_API_TYPE_PREFIX,
            PT_API_TYPE_NPU: NPU_API_TYPE_PREFIX,
            PT_API_TYPE_ATEN: ATEN_API_TYPE_PREFIX,
            PT_API_TYPE_DIST: DIST_API_TYPE_PREFIX,
            PT_API_TYPE_DIST_C10D: DIST_API_TYPE_PREFIX,
            PT_API_TYPE_NPU_DIST: DIST_API_TYPE_PREFIX,
            PT_API_TYPE_MINDSPEED: MINDSPEED_API_TYPE_PREFIX
        },
        MS_FRAMEWORK: {
            MS_API_TYPE_OPS: FUNCTIONAL_API_TYPE_PREFIX,
            MS_API_TYPE_TENSOR: TENSOR_API_TYPE_PREFIX,
            MS_API_TYPE_STUB_TENSOR: TENSOR_API_TYPE_PREFIX,
            MS_API_TYPE_MINT: MINT_API_TYPE_PREFIX,
            MS_API_TYPE_MINT_FUNC: MINT_FUNC_API_TYPE_PREFIX,
            MS_API_TYPE_COM: DIST_API_TYPE_PREFIX,
            MS_API_TYPE_MINT_DIST: MINT_DIST_API_TYPE_PREFIX
        },
        MT_FRAMEWORK: {
            PT_API_TYPE_FUNCTIONAL: FUNCTIONAL_API_TYPE_PREFIX,
            PT_API_TYPE_TENSOR: TENSOR_API_TYPE_PREFIX,
            PT_API_TYPE_TORCH: TORCH_API_TYPE_PREFIX,
            PT_API_TYPE_NPU: NPU_API_TYPE_PREFIX,
            PT_API_TYPE_DIST: DIST_API_TYPE_PREFIX
        }
    }

    def _fused_adamw_(
            self,
            grads,
            exp_avgs,
            exp_avg_sqs,
            max_exp_avg_sqs,
            state_steps,
            *,
            lr,
            beta1,
            beta2,
            weight_decay,
            eps,
            amsgrad,
            maximize,
            grad_scale=None,
            found_inf=None
    ):
        pass

    API_WITH_SELF_ARG = {
        'Torch._fused_adamw_': _fused_adamw_
    }

    ASCEND = "ASCEND"
    MATCH_MODE_NAME = "pure name"
    MATCH_MODE_MAPPING = "mapping"
    MATCH_MODE_SIMILARITY = "similarity"
    CONFIG_CHECK_PASS = "pass"
    CONFIG_CHECK_WARNING = "warning"
    CONFIG_CHECK_ERROR = "error"

    MIX_DUMP_NAMES = {'graph', 'pynative'}

    MEGATRON_MICRO_STEP_NUMBER = 'megatron_micro_step_number'

    TORCHAIR_FX_DATA_DIRECTORY = 'msprobe_fx_dump'
    TORCHAIR_GE_DATA_DIRECTORY = 'msprobe_ge_dump'

    FULL_OP_NAME = "full_op_name"
    DIRECTION = "direction"
    CALL_DIRECTION = "call_direction"
    OP_NO_NUMBER = "op_no_number"
    BACKWARD_CALL_ORDER = "backward_call_order"
    SUFFIX = "suffix"


class Data2DBConst:
    BATCH_SIZE = 5000
    FORWARD = "forward"
    BACKWARD = "backward"
    RECOMPUTE = "recompute"
    PARAMETERS_GRAD = "parameters_grad"
    METRICS = [FORWARD, BACKWARD, RECOMPUTE, PARAMETERS_GRAD]
    DEFAULT_TAGS = ["Module", "Cell"]
    ORDERED_STAT = ["norm", "max", "min", "mean"]
    TAG_DEFAULT = "default"
    TAG_INDEX = "index"
    TAG_LAYER = "layer"
    TAG_MODULE = "module"
    TAG_FUNCTION = "function"
    MAX_PARTITION = 10_000_000
    MIN_PARTITION = 10
    MAX_MICRO_STEP = 10000
    MIN_MICRO_STEP = 1
    SUPPORT_TYPE = ["torch.Tensor", "mindspore.Tensor", "mindtorch.Tensor"]
    SUPPORT_DTYPE = [Const.FLOAT16,
                     Const.FLOAT32,
                     Const.BFLOAT16,
                     Const.TORCH_FLOAT16,
                     Const.TORCH_FLOAT32,
                     Const.TORCH_BFLOAT16]


class CompareConst:
    """
    Class for compare module const
    """
    SPACE = " "
    NAME = "Name"
    # compare result column name
    NPU_NAME = "NPU Name"
    BENCH_NAME = "Bench Name"
    NPU_DTYPE = "NPU Dtype"
    BENCH_DTYPE = "Bench Dtype"
    NPU_SHAPE = "NPU Tensor Shape"
    BENCH_SHAPE = "Bench Tensor Shape"
    NPU_CSV_FILE = "NPU CSV File"
    BENCH_CSV_FILE = "Bench CSV File"
    NPU_MAX = "NPU max"
    NPU_MIN = "NPU min"
    NPU_MEAN = "NPU mean"
    NPU_NORM = "NPU l2norm"
    NPU_P2POP_PEER = "NPU P2POp peer"

    BENCH_MAX = "Bench max"
    BENCH_MIN = "Bench min"
    BENCH_MEAN = "Bench mean"
    BENCH_NORM = "Bench l2norm"
    MAX_DIFF = "Max diff"
    MIN_DIFF = "Min diff"
    MEAN_DIFF = "Mean diff"
    NORM_DIFF = "L2norm diff"
    COSINE = "Cosine"
    EUC_DIST = "EucDist"
    MAX_ABS_ERR = "MaxAbsErr"
    MAX_RELATIVE_ERR = "MaxRelativeErr"
    MIN_RELATIVE_ERR = "MinRelativeErr"
    MEAN_RELATIVE_ERR = "MeanRelativeErr"
    NORM_RELATIVE_ERR = "NormRelativeErr"
    REQ_GRAD_CONSIST = "Requires_grad Consistent"
    NPU_REQ_GRAD = "NPU Requires_grad"
    BENCH_REQ_GRAD = "Bench Requires_grad"
    ACCURACY = "Accuracy Reached or Not"
    STACK = "NPU_Stack_Info"
    DATA_NAME = "Data_name"
    ERROR_MESSAGE = "Err_message"
    ONE_THOUSANDTH_ERR_RATIO = "One Thousandth Err Ratio"
    FIVE_THOUSANDTHS_ERR_RATIO = "Five Thousandths Err Ratio"
    NPU_MD5 = "NPU MD5"
    BENCH_MD5 = "BENCH MD5"
    RESULT = "Result"
    MAGNITUDE = 0.5
    OP_NAME = "op_name"
    OP_NAME_UPDATE = "op_name_update"
    STRUCT = "struct"
    INPUT_STRUCT = "input_struct"
    KWARGS_STRUCT = "kwargs_struct"
    OUTPUT_STRUCT = "output_struct"
    PARAMS_STRUCT = "params_struct"
    PARAMS_GRAD_STRUCT = "params_grad_struct"
    DEBUG_STRUCT = "debug_struct"
    SUMMARY = "summary"
    COMPARE_RESULT = "compare_result"
    COMPARE_MESSAGE = "compare_message"
    MAX_EXCEL_LENGTH = 1048500
    YES = "Yes"
    NO = "No"
    STATISTICS_INDICATOR_NUM = 4
    EPSILON = 1e-10
    COMPARE_ENDS_SUCCESSFULLY = "msprobe compare ends successfully."
    DEFAULT_RATIO_VALUE = 10000
    THOUSANDTH_PASS_VALUE = 0.999
    ZERO_SHAPE = '(0,)'

    BENCHMARK_COMPARE_ALGORITHM_NAME = "标杆比对法"
    ULP_COMPARE_ALGORITHM_NAME = "ULP误差比对法"
    BINARY_CONSISTENCY_ALGORITHM_NAME = "二进制一致法"
    ABSOLUTE_THRESHOLD_ALGORITHM_NAME = "绝对阈值法"
    THOUSANDTH_STANDARD_ALGORITHM_NAME = "双千指标法"
    ACCUMULATIVE_ERROR_COMPARE_ALGORITHM_NAME = "累积误差比对法"

    ABSOLUTE_THRESHOLD = 'absolute_threshold'
    BINARY_CONSISTENCY = 'binary_consistency'
    ULP_COMPARE = 'ulp_compare'
    THOUSANDTH_STANDARD = 'thousandth_threshold'
    BENCHMARK = 'benchmark'
    ACCUMULATIVE_ERROR_COMPARE = 'accumulative_error_compare'

    SMALL_VALUE_ERR_RATIO = "small_value_err_ratio"
    RMSE_RATIO = "rmse_ratio"
    MAX_REL_ERR_RATIO = "max_rel_err_ratio"
    MEAN_REL_ERR_RATIO = "mean_rel_err_ratio"
    EB_RATIO = "eb_ratio"

    SMALL_VALUE = "small_value"
    RMSE = "rmse"
    MAX_REL_ERR = "max_rel_err"
    MEAN_REL_ERR = "mean_rel_err"
    EB = "eb"

    SMALL_VALUE_ERR_STATUS = "small_value_err_status"
    RMSE_STATUS = "rmse_status"
    MAX_REL_ERR_STATUS = "max_rel_err_status"
    MEAN_REL_ERR_STATUS = "mean_rel_err_status"
    EB_STATUS = "eb_status"

    MEAN_ULP_ERR = "mean_ulp_err"
    ULP_ERR_PROPORTION = "ulp_err_proportion"
    ULP_ERR_PROPORTION_RATIO = "ulp_err_proportion_ratio"

    ULP_ERR_STATUS = "ulp_err_status"

    ALL_COMPARE_INDEX = [COSINE, EUC_DIST, MAX_ABS_ERR, MAX_RELATIVE_ERR,
                         ONE_THOUSANDTH_ERR_RATIO, FIVE_THOUSANDTHS_ERR_RATIO]
    SUMMARY_COMPARE_INDEX = [MAX_DIFF, MIN_DIFF, MEAN_DIFF, NORM_DIFF,
                             MAX_RELATIVE_ERR, MIN_RELATIVE_ERR, MEAN_RELATIVE_ERR, NORM_RELATIVE_ERR]
    MD5_COMPARE_INDEX = [RESULT]

    BASIC_INFO = [NPU_NAME, BENCH_NAME, NPU_DTYPE, BENCH_DTYPE, NPU_SHAPE, BENCH_SHAPE, NPU_REQ_GRAD, BENCH_REQ_GRAD]
    SUMMARY_INFO = [NPU_MAX, NPU_MIN, NPU_MEAN, NPU_NORM, BENCH_MAX, BENCH_MIN, BENCH_MEAN, BENCH_NORM]

    COMPARE_RESULT_HEADER = BASIC_INFO + ALL_COMPARE_INDEX + SUMMARY_INFO + [REQ_GRAD_CONSIST, RESULT, ERROR_MESSAGE]

    SUMMARY_COMPARE_RESULT_HEADER = BASIC_INFO + SUMMARY_COMPARE_INDEX + SUMMARY_INFO + [REQ_GRAD_CONSIST, RESULT,
                                                                                         ERROR_MESSAGE]

    MD5_COMPARE_RESULT_HEADER = BASIC_INFO + [NPU_MD5, BENCH_MD5, REQ_GRAD_CONSIST] + [RESULT, ERROR_MESSAGE]

    COMPARE_RESULT_HEADER_STACK = COMPARE_RESULT_HEADER + [STACK]

    SUMMARY_COMPARE_RESULT_HEADER_STACK = SUMMARY_COMPARE_RESULT_HEADER + [STACK]

    MD5_COMPARE_RESULT_HEADER_STACK = MD5_COMPARE_RESULT_HEADER + [STACK]

    HEAD_OF_COMPARE_MODE = {
        Const.ALL: COMPARE_RESULT_HEADER,
        Const.SUMMARY: SUMMARY_COMPARE_RESULT_HEADER,
        Const.MD5: MD5_COMPARE_RESULT_HEADER
    }

    # dtype match

    DTYPE_MATCH_GROUPS = [
        {Const.FLOAT16, Const.FLOAT32, Const.BFLOAT16},
        {Const.TORCH_FLOAT16, Const.TORCH_FLOAT32, Const.TORCH_BFLOAT16}
    ]

    # read_op
    IO_NAME_MAPPING = {
        Const.INPUT_ARGS: '.input',
        Const.INPUT_KWARGS: '.input',
        Const.INPUT: '.input',
        Const.OUTPUT: '.output',
        Const.PARAMS: '.parameters'
    }

    # state to struct mapping
    STATE_TO_STRUCT_MAPPING = {
        Const.INPUT: INPUT_STRUCT,
        Const.KWARGS: INPUT_STRUCT,
        Const.OUTPUT: OUTPUT_STRUCT,
        Const.PARAMS: PARAMS_STRUCT,
        Const.PARAMS_GRAD: PARAMS_GRAD_STRUCT,
        Const.DEBUG: DEBUG_STRUCT
    }

    # compare standard
    HUNDRED_RATIO_THRESHOLD = 0.01
    THOUSAND_RATIO_THRESHOLD = 0.001
    FIVE_THOUSAND_RATIO_THRESHOLD = 0.005
    TEN_THOUSAND_RATIO_THRESHOLD = 0.0001
    COSINE_THRESHOLD = 0.9999
    ULP_FLOAT32_THRESHOLD = 32
    ULP_FLOAT16_THRESHOLD = 1

    # compare result data
    NO_REAL_DATA = 'No real data'
    API_UNMATCH = 'api unmatched'
    READ_NONE = 'No data'
    NONE = 'None'
    SHAPE_UNMATCH = 'shape unmatched'
    DIFF = 'Different'
    DIFF_FLAG = 'Diff'
    UNSUPPORTED = 'unsupported'
    NAN = 'Nan'
    PASS = 'pass'
    WARNING = 'warning'
    ERROR = 'error'
    TRUE = 'TRUE'
    FALSE = 'FALSE'
    SKIP = 'SKIP'
    N_A = 'N/A'
    INF = 'inf'
    NEG_INF = '-inf'
    BFLOAT16_MIN = -3.3895313892515355e+38
    BFLOAT16_MAX = 3.3895313892515355e+38
    BFLOAT16_EPS = 3.90625e-3  # 2 ** -8

    # accuracy standards
    COS_THRESHOLD = 0.99
    MAX_ABS_ERR_THRESHOLD = 0.001
    MAX_RELATIVE_ERR_THRESHOLD = 0.001
    COS_MAX_THRESHOLD = 0.9
    MAX_ABS_ERR_MAX_THRESHOLD = 1
    ACCURACY_CHECK_YES = "Yes"
    ACCURACY_CHECK_NO = "No"
    ACCURACY_CHECK_UNMATCH = "Unmatched"

    # error message
    NO_BENCH = "No bench data matched."

    # compare const
    FLOAT_TYPE = [np.half, np.single, float, np.double, np.float64, np.longdouble]

    # acc_check const
    MAX_TOKENS = 65536
    SPECIAL_SPARSE_MOED = 4

    OVERFLOW_LIST = ['nan\t', 'inf\t', '-inf\t', 'nan', 'inf', '-inf']

    MS_GRAPH_BASE = {
        NPU_NAME: None, BENCH_NAME: None, NPU_DTYPE: None, BENCH_DTYPE: None, NPU_SHAPE: None, BENCH_SHAPE: None,
        NPU_MAX: None, NPU_MIN: None, NPU_MEAN: None, NPU_NORM: None, BENCH_MAX: None, BENCH_MIN: None,
        BENCH_MEAN: None, BENCH_NORM: None, ACCURACY: '', ERROR_MESSAGE: ''
    }
    MS_GRAPH_NPY = {
        COSINE: None, EUC_DIST: None, MAX_ABS_ERR: None, MAX_RELATIVE_ERR: None, ONE_THOUSANDTH_ERR_RATIO: None,
        FIVE_THOUSANDTHS_ERR_RATIO: None
    }
    MS_GRAPH_STATISTIC = {
        MAX_DIFF: None, MIN_DIFF: None, MEAN_DIFF: None, NORM_DIFF: None, MAX_RELATIVE_ERR: None,
        MIN_RELATIVE_ERR: None, MEAN_RELATIVE_ERR: None, NORM_RELATIVE_ERR: None
    }
    MS_GRAPH_CSV = {
        NPU_CSV_FILE: None, BENCH_CSV_FILE: None
    }

    API_MAPPING_KEYS_TO_COMPARE = [
        ('ms_args', 'pt_args'),
        ('ms_outputs', 'pt_outputs'),
        ('ms_parameters', 'pt_parameters'),
        ('ms_parameters_grad', 'pt_parameters_grad')
    ]

    INPUT_PATTERN = Const.SEP + Const.INPUT + Const.SEP
    KWARGS_PATTERN = Const.SEP + Const.KWARGS + Const.SEP
    OUTPUT_PATTERN = Const.SEP + Const.OUTPUT + Const.SEP
    PARAMS_PATTERN = Const.SEP + Const.PARAMS + Const.SEP
    PARAMS_GRAD_PATTERN = Const.SEP + Const.PARAMS_GRAD + Const.SEP

    CMP_KEY = 'compare_key'
    CMP_SHAPE = 'compare_shape'

    OP_NAME_X = 'op_name_x'
    MATCH_RESULT_COLUMNS = [
        OP_NAME_X, 'dtype_x', 'shape_x', 'summary_x', 'stack_info_x', 'state_x', 'api_origin_name_x',
        'requires_grad_x', 'data_name_x',
        CMP_KEY, CMP_SHAPE,
        'op_name_y', 'dtype_y', 'shape_y', 'summary_y', 'stack_info_y', 'state_y', 'api_origin_name_y',
        'requires_grad_y', 'data_name_y'
    ]

    INTERNAL_API_MAPPING_FILE = 'ms_to_pt_api.yaml'
    UNREADABLE = 'unreadable data'
    NPU_DUMP_DATA_DIR = 'npu_dump_data_dir'
    BENCH_DUMP_DATA_DIR = 'bench_dump_data_dir'
    NO_REAL_DATA_FLAG = '-1'

    # ATB data comparison
    STATS_COMPARISION_MODE = 'stats'
    TENSOR_COMPARISION_MODE = 'tensor'
    UNKNOWN_COMPARISION_MODE = "unknown"
    ATB_DATA_NAME = 'Data Name'
    DEVICE_AND_PID = 'Device and PID'
    EXECUTION_COUNT = 'Execution Count'
    DATA_TYPE = 'Data Type'
    DATA_SHAPE = 'Data Shape'
    TENSOR_PATH = "tensor_path"


class FileCheckConst:
    """
    Class for file check const
    """
    READ_ABLE = "r"
    WRITE_ABLE = "w"
    EXECUTE_ABLE = "x"
    READ_WRITE_ABLE = "rw"
    READ_EXECUTE_ABLE = "rx"
    WRITE_EXECUTE_ABLE = "wx"
    READ_WRITE_EXECUTE_ABLE = "rwx"
    PERM_OPTIONS = [
        READ_ABLE,
        WRITE_ABLE,
        EXECUTE_ABLE,
        READ_WRITE_ABLE,
        READ_EXECUTE_ABLE,
        WRITE_EXECUTE_ABLE,
        READ_WRITE_EXECUTE_ABLE
    ]

    DIRECTORY_LENGTH = 4096
    FILE_NAME_LENGTH = 255
    FILE_VALID_PATTERN = r"^[a-zA-Z0-9_.:/-]+$"
    FILE_PATTERN = r'^[a-zA-Z0-9_./-]+$'
    PKL_SUFFIX = ".pkl"
    NUMPY_SUFFIX = ".npy"
    JSON_SUFFIX = ".json"
    PT_SUFFIX = ".pt"
    CSV_SUFFIX = ".csv"
    XLSX_SUFFIX = ".xlsx"
    YAML_SUFFIX = ".yaml"
    IR_SUFFIX = ".ir"
    ZIP_SUFFIX = ".zip"
    SHELL_SUFFIX = ".sh"
    LOG_SUFFIX = ".log"
    DB_SUFFIX = '.db'
    ONNX_SUFFIX = '.onnx'
    OM_SUFFIX = '.om'
    BIN_SUFFIX = '.bin'
    MAX_BIN_SIZE = 10737418240          # 10G
    MAX_COMMON_FILE_SIZE = 10737418240  # 10G
    MAX_PKL_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_NUMPY_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_JSON_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_PT_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_CSV_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_XLSX_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_YAML_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_IR_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_ZIP_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_FILE_IN_ZIP_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_FILE_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    COMMON_FILE_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_LOG_SIZE = 10737418240  # 1 * 1024 * 1024 * 1024
    MAX_DB_SIZE = 107374182400  # 100 * 1024 * 1024 * 1024
    MAX_ONNX_SIZE = 10737418240  # 1 * 1024 * 1024 * 1024
    MAX_OM_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    DIR = "dir"
    FILE = "file"
    DATA_DIR_AUTHORITY = 0o750
    DATA_FILE_AUTHORITY = 0o640
    FILE_SIZE_DICT = {
        PKL_SUFFIX: MAX_PKL_SIZE,
        NUMPY_SUFFIX: MAX_NUMPY_SIZE,
        JSON_SUFFIX: MAX_JSON_SIZE,
        PT_SUFFIX: MAX_PT_SIZE,
        CSV_SUFFIX: MAX_CSV_SIZE,
        XLSX_SUFFIX: MAX_XLSX_SIZE,
        YAML_SUFFIX: MAX_YAML_SIZE,
        IR_SUFFIX: MAX_IR_SIZE,
        ZIP_SUFFIX: MAX_ZIP_SIZE,
        LOG_SUFFIX: MAX_LOG_SIZE,
        DB_SUFFIX: MAX_DB_SIZE,
        ONNX_SUFFIX: MAX_ONNX_SIZE,
        OM_SUFFIX: MAX_OM_SIZE,
        BIN_SUFFIX: MAX_BIN_SIZE
    }
    CSV_BLACK_LIST = r'^[＋－＝％＠\+\-=%@]|;[＋－＝％＠\+\-=%@]'


class OverflowConst:
    """
    Class for Overflow
    """
    OVERFLOW_ORIGINAL_MODE = 0
    OVERFLOW_DEBUG_MODE = 1


class MsgConst:
    """
    Class for log messages const
    """
    MSPROBE_LOG_LEVEL = "MSPROBE_LOG_LEVEL"
    LOG_LEVEL_ENUM = ["0", "1", "2", "3", "4"]
    LOG_LEVEL = ["DEBUG", "INFO", "WARNING", "ERROR"]

    class LogLevel:
        class DEBUG:
            value = 0

        class INFO:
            value = 1

        class WARNING:
            value = 2

        class ERROR:
            value = 3

    SPECIAL_CHAR = ["\n", "\r", "\u007F", "\b", "\f", "\t", "\u000B", "%08", "%0a", "%0b", "%0c", "%0d", "%7f"]

    NOT_CREATED_INSTANCE = "PrecisionDebugger instance is not created."


class GraphMode:
    NPY_MODE = "NPY_MODE"
    STATISTIC_MODE = "STATISTIC_MODE"
    ERROR_MODE = "ERROR_MODE"


class MonitorConst:
    """
    Class for monitor const
    """

    # monitor config set default values
    DEFAULT_GRAD_ACC_STEPS = 1
    DEFAULT_START_ITERATION = 0
    DEFAULT_START_STEP = 0
    DEFAULT_MAX_COLLECT_TIMES = 1e8
    DEFAULT_MIN_COLLECT_TIMES = 0
    DEFAULT_STEP_INTERVAL = 1

    OP_LIST = ["norm", "min", "max", "zeros", "nans", "id", "mean", "shape", "dtype"]
    OP_MONVIS_SUPPORTED = [
        "norm", "min", "max", "zeros", "nans", "mean",
        "entropy", "softmax_max", "sr", "kernel_norm", "std_x", "jacobian",
        "proxy", "token_similarity"
    ]
    METRICS_TRENDVIS_SUPPORTED = [
        "actv", "actv_grad", "exp_avg", "exp_avg_sq",
        "grad_unreduced", "grad_reduced", "param_origin", "param_updated", "other"
    ]
    MONITOR_OUTPUT_DIR = "MONITOR_OUTPUT_DIR"
    DEFAULT_MONITOR_OUTPUT_DIR = "./monitor_output"
    DATABASE = "database"
    EMAIL = "email"
    OPT_TY = ['Megatron_DistributedOptimizer', 'Megatron_Float16OptimizerWithFloat16Params']
    DEEPSPEED_OPT_TY = (
        "DeepSpeedZeroOptimizer_Stage0",
        "DeepSpeedZeroOptimizer_Stage1_or_2",
        "DeepSpeedZeroOptimizer_Stage3"
    )
    DEEPSPEED_ZERO_OPT_FILTER = "DeepSpeedZeroOptimizer"
    RULE_NAME = ['AnomalyTurbulence', 'AnomalyNan']
    L2_HOOKS = ["linear_hook", "attention_hook"]
    SA_ORDERS = ["s,b,h,d", "b,s,h,d"]

    SLICE_SIZE = 20480
    # used for name
    DOT = "."
    NAME_SEP = ":"
    INPUT_GRAD = "input_grad"
    OUTPUT_GRAD = "output_grad"
    ACTV_IN = "input"
    ACTV_OUT = "output"
    ACTVGRAD_IN = "input_grad"
    ACTVGRAD_OUT = "output_grad"
    FSDP_FLAT_SEP = "_fsdp_wrapped_module."
    # used for tasks
    ACTV = "actv"
    ACTVGRAD = "actv_grad"
    POST_GRAD = "post_grad"
    PRE_GRAD = "pre_grad"
    PRE_PARAM = "param_origin"
    POST_PARAM = "param_updated"
    ACC_GRAD = "acc_grad"
    PREFIX_POST = "post"
    PREFIX_PRE = "pre"
    EXP_AVG = "exp_avg"
    EXP_AVG_SQ = "exp_avg_sq"

    CSV_HEADER = ["vpp_stage", "name", "step"]
    CSV_HEADER_MICRO_STEP = ["vpp_stage", "name", "step", "micro_step"]
    OUTPUT_DIR_PATTERN = r"([\w-]{0,20})-rank(\d{1,5})-"
    ANOMALY_JSON = "anomaly.json"
    ANALYSE_JSON = "anomaly_analyse.json"
    TENSORBOARD = "tensorboard"
    CSV = "csv"
    API = "api"
    HEADER_NAME = 'name'
    MAX_NDIGITS = 20

    DEFAULT_STAGE = -1
    FORWARD_STAGE = 0
    BACKWARD_STAGE = 1
    OPTIMIZER_STAGE = 2
    FORWARD_KEY = [ACTV]
    BACKWARD_KEY = [ACTVGRAD, PRE_GRAD, POST_GRAD, ACC_GRAD]
    OPTIMIZER_KEY = [EXP_AVG, EXP_AVG_SQ]

    TRAIN_STAGE = {}
    for key in FORWARD_KEY:
        TRAIN_STAGE[key] = FORWARD_STAGE
    for key in BACKWARD_KEY:
        TRAIN_STAGE[key] = BACKWARD_STAGE
    for key in OPTIMIZER_KEY:
        TRAIN_STAGE[key] = OPTIMIZER_STAGE

    # csv2db
    DEFAULT_INT_VALUE = 0
    MAX_PROCESS_NUM = 128
    CSV_FILE_PATTERN = r"_(\d+)-(\d+)\.csv"
    BATCH_SIZE = 10000
    MAX_PARTITION = 10_000_000
    MIN_PARTITION = 10
    


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.exceptions import ParseJsonException
from msprobe.core.common.file_utils import load_json
from msprobe.core.common.log import logger


def parse_json_info_forward_backward(json_path):
    dump_json = load_json(json_path)

    real_data_path = dump_json.get("dump_data_dir")
    dump_data = dump_json.get("data")
    if dump_data is None:
        raise ParseJsonException(ParseJsonException.InvalidDumpJson, 
                                 "something wrong with dump, no data found in dump.json")
    if not dump_data:
        logger.warning("data field is empty, no overflow data found.")

    forward_data = {}
    backward_data = {}
    for data_name, data_item in dump_data.items():
        if "Module" in data_name:
            continue
        if "forward" in data_name:
            api_name = parse_data_name_with_pattern(data_name, "forward", json_path)
            forward_data.update({api_name: data_item})
        elif "backward" in data_name:
            api_name = parse_data_name_with_pattern(data_name, "backward", json_path)
            backward_data.update({api_name: data_item})
        else:
            raise ParseJsonException(ParseJsonException.UnexpectedNameStruct,
                                     f"{data_name} in file {json_path}.")

    return forward_data, backward_data, real_data_path


def parse_data_name_with_pattern(data_name, pattern, json_path):
    name_struct = data_name.split('.')
    if not name_struct[-1] == pattern:
        raise ParseJsonException(ParseJsonException.UnexpectedNameStruct, f"{data_name} in file {json_path}")
    api_name = '.'.join(name_struct[:-1])
    return api_name



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import json
import os
import uuid
import importlib
from collections import defaultdict
from datetime import datetime
from functools import partial
from itertools import cycle

import pytz
import torch
import torch.distributed as dist
import pandas as pd
from torch.utils.hooks import BackwardHook
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP

from msprobe.core.common.const import MonitorConst, Const
from msprobe.core.common.file_utils import load_json, save_json, create_directory
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.monitor.anomaly_processor import AnomalyScanner, AnomalyDataFactory, AnomalyDataWriter
from msprobe.core.common.file_utils import write_df_to_csv
from msprobe.core.common.utils import analyze_api_call_stack
from msprobe.core.monitor.utils import validate_config, validate_ops, \
    get_output_base_dir, get_target_output_dir, chmod_tensorboard_dir, validate_set_monitor
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import is_recomputation
from msprobe.pytorch.monitor.utils import get_param_struct
from msprobe.pytorch.monitor.data_writers import SummaryWriterWithAD, CSVWriterWithAD, BaseWriterWithAD, WriterInput
from msprobe.pytorch.monitor.distributed.wrap_distributed import api_register, create_hooks, op_aggregate, \
    get_process_group
from msprobe.pytorch.monitor.features import get_sign_matches, cal_qkt
from msprobe.pytorch.monitor.module_metric import get_metrics, get_summary_writer_tag_name, \
    TensorMetrics, squash_param_name, get_entropy_metric, get_sr_metric
from msprobe.pytorch.monitor.optimizer_collect import OptimizerMonFactory
from msprobe.pytorch.monitor.visualizer import HeatmapVisualizer

torch_version_above_or_equal_2 = torch.__version__.split('+')[0] >= '2.0'
if not torch_version_above_or_equal_2:
    raise ValueError("monitor require torch>=2.0")

FORMAT_MAPPING = {
    MonitorConst.TENSORBOARD: SummaryWriterWithAD,
    MonitorConst.CSV: CSVWriterWithAD,
    MonitorConst.API: BaseWriterWithAD
}
start_step = 0


def param_is_not_tensor_parallel_duplicate(param, tp_group):
    return (hasattr(param, 'tensor_model_parallel') and param.tensor_model_parallel) or (
            torch.distributed.get_rank(group=tp_group) == 0
    )


def param_is_data_parallel_duplicate(dp_group):
    return torch.distributed.get_rank(group=dp_group) != 0


class ModuleHookContext:
    def __init__(self, module_name) -> None:
        self.micro_step = 0
        self.actv = defaultdict(dict)
        self.actvgrad = []
        self.module_name = module_name
        self.struct = {}
        self.stack = ""

    def reset(self):
        self.actv.clear()
        self.actvgrad.clear()


class FeatureHookContext:
    def __init__(self, module_name):
        self.step = 0
        self.micro_step = 0
        self.attention_feature = {}
        self.linear_feature = {}
        self.module_name = module_name

    def reset(self):
        self.attention_feature.clear()
        self.linear_feature.clear()


class OptimizerContext:
    def __init__(self) -> None:
        self.step = start_step
        self.param_mg_direction = defaultdict(float)
        self.param_adam_update = defaultdict()
        self.param_adam_ratio = defaultdict()
        self.param_weight_grad = defaultdict()
        self.param_exp_avg = defaultdict()
        self.exp_avg_metric = {}
        self.param_exp_avg_sq = defaultdict()
        self.exp_avg_sq_metric = {}
        self.metric_dict = {}
        self.param_metric = {}

    def reset(self):
        self.param_mg_direction.clear()
        self.param_adam_update.clear()
        self.param_adam_ratio.clear()
        self.param_weight_grad.clear()
        self.param_exp_avg.clear()
        self.exp_avg_metric.clear()
        self.param_exp_avg_sq.clear()
        self.exp_avg_sq_metric.clear()
        self.metric_dict.clear()
        self.param_metric.clear()


class CommunicationContext:
    def __init__(self) -> None:
        self.data = {}

    @staticmethod
    def _agg(data):
        aggregated_data = {}
        for tag, op2tensorlist in data.items():
            aggregated_data[tag] = {}
            for op, tensorlist in op2tensorlist.items():
                aggregated_data[tag][op] = op_aggregate(op, tensorlist)
        return aggregated_data

    def reset(self):
        self.data = {}

    def aggregate(self):
        self.data = self._agg(self.data)


class GradContext:
    def __init__(self) -> None:
        self.pre = {}
        self.post = {}
        self.actv = {}

    def reset(self):
        self.pre.clear()
        self.post.clear()
        self.actv.clear()


class TrainerMon:
    tensor_metrics = TensorMetrics()

    # 保留原opt_ty参数, 兼容msprobe1.2.2前旧版本
    def __init__(self, config_file_path, process_group=None, params_have_main_grad=True, opt_ty=None) -> None:
        # TYPE1: 只在这里初始化的变量, 不会随着训练中途config配置改变而重置
        self.config_file_path = config_file_path
        self.process_group = get_process_group(process_group)
        self.params_have_main_grad = params_have_main_grad
        self.update_heatmap_visualizer = defaultdict(HeatmapVisualizer)
        self.ratio_heatmap_visualizer = defaultdict(HeatmapVisualizer)
        self.fsdp_post_backward_hook = None
        self.fsdp2_foreach_reduce = None
        self.config_timestamp = 0  # 后面有校验时间戳, 首次监测无需为了更新config文件时间戳而去改, 可通过dynamic_on开关直接打开
        self.config = load_json(config_file_path)
        validate_config(self.config)

        self.squash_name = self.config.get('squash_name', True)  # 不允许修改防止前后名字对不上
        local_tz = pytz.timezone("Asia/Shanghai")  # 根据需要调整为目标时区
        cur_time = datetime.now(local_tz).strftime('%b%d_%H-%M-%S')
        self.unique_id = str(uuid.uuid4())[:8]
        self.output_base_dir = get_output_base_dir()
        time_tags = self.config.get("append_output", [])
        if dist.is_initialized():
            self.rank = dist.get_rank()
            if time_tags:
                output_append_dirs = get_target_output_dir(self.output_base_dir, time_tags[0], time_tags[1])
                if str(self.rank) in output_append_dirs:
                    self.tensorboard_dir = output_append_dirs[str(self.rank)]
                    logger.info(f"append rank({self.rank}) result to {self.tensorboard_dir}")
            else:
                self.tensorboard_dir = os.path.join(self.output_base_dir,
                                                    f"{cur_time}-rank{self.rank}-{self.unique_id}")
            self.pp_stage = dist.get_group_rank(self.process_group, self.rank)
            self.group_mates = dist.get_process_group_ranks(self.process_group)
        else:
            self.rank = 0
            self.tensorboard_dir = os.path.join(self.output_base_dir, f"{cur_time}-rank{self.rank}-{self.unique_id}")
            self.pp_stage = 0
            self.group_mates = [0]

        # TYPE2: 只会在set_monitor()主调中赋值的变量
        self.model = None
        self.vpp = False
        self.dp_group = None
        self.tp_group = None
        self.enable_megatron = False
        self.enable_deepspeed = False
        self.fsdp_wrapped_module = False
        self.fsdp2_wrapped_module = False
        self.micro_batch_number = 1
        self.optimizer_mon = None
        self.optimizer_trans = None

        # TYPE3: 会随着训练中途config配置更新或监测状态改变而重置的变量
        self.module_fwd_hook_context_by_module = defaultdict(ModuleHookContext)
        self.module_bwd_hook_context_by_module = defaultdict(ModuleHookContext)
        self.feature_hook_context_by_module = defaultdict(FeatureHookContext)
        self.optimizer_context = defaultdict(OptimizerContext)
        self.cc_context = defaultdict(CommunicationContext)
        self.grad_context = GradContext()
        self.handles = defaultdict(list)
        self.param2name = defaultdict(str)
        self.name2indices = defaultdict()
        self.name2param = {}
        self.origin2squash = {}
        self.duplicate_param = {}
        self.name2tag = {}
        self.param_name_call_id = {}
        self.fsdp_param_name_map = {}
        self.call_id = 0
        self.module_struct = defaultdict(dict)
        self.grad_accs = []
        self.weight_hooked = False
        self.optimizer_hooked = False
        self.param_registered = False
        self.struct_printed = False
        self.pre_step_hooks = []
        self.post_step_hooks = []

        # 动静态区分
        self.dynamic_enable = os.getenv("DYNAMIC_MONITOR", 'False').lower() == 'true'
        if self.dynamic_enable:
            logger.warning(f"DYNAMIC_MONITOR is set, "
                           f"please make sure you have 'dynamic_on' and 'collect_times' in {self.config_file_path}")
            self.monitoring = False
        else:
            self.set_config()
            # 静态且collect_times>0时在第0步self.monitoring就可以True, 动态默认在下一步开启
            if self.collect_times > 0:
                self.monitoring = True

    def __del__(self):
        if hasattr(self, "summary_writer"):
            self.summary_writer.close()

    @property
    def ops(self):
        return self._ops

    @ops.setter
    def ops(self, value):
        self._ops = validate_ops(value)

    @staticmethod
    def has_register_backward_hook(module_name, module):
        if hasattr(module, '_backward_hooks') and \
                len(module._backward_hooks) > 0 and \
                module._is_full_backward_hook is False:
            logger.warning(
                f"The {module_name} has registered deprecated register_backward_hook,"
                f"which may cause abnormal data dump. The backward input/output for this module will be skipped."
            )
            return True
        return False

    @staticmethod
    def generate_cc_metrics(cc_name, cc_tensor):
        metrics = defaultdict(dict)
        rank = dist.get_rank() if dist.is_initialized() else None
        for op, tag2tensor in cc_tensor.data.items():
            for tag, tensor in tag2tensor.items():
                key = get_summary_writer_tag_name(cc_name, tag, rank)
                metrics[op].update({key: tensor})
        cc_tensor.reset()
        return metrics

    @staticmethod
    def get_linear_hook_target(module):
        if isinstance(module, torch.nn.Embedding):
            return ''
        if hasattr(module, "num_embeddings") or hasattr(module, "vocab_start_index"):
            return ''
        for weight_name in ["weight", "wg"]:
            if hasattr(module, weight_name) and isinstance(getattr(module, weight_name), torch.Tensor):
                if getattr(module, weight_name).dim() == 2:
                    return weight_name
        return ''

    def set_config(self):
        logger.info(f"current config: {self.config}")
        self.start_step = self.config.get("start_step", 0)
        self.collect_times = self.config.get("collect_times", 100000000)  # 默认大值, 目的是一直采集
        self.step_interval = self.config.get("step_interval", 1)
        self.has_collect_times = 0  # 重设采集计数器
        self.print_struct = self.config.get("print_struct", False)
        self.module_rank_list = self.config.get("module_ranks", [])
        self.format = self.config.get('format', MonitorConst.CSV)
        self.eps = self.config.get('eps', 1e-8)
        self.ops = self.config.get('ops', [])
        self.ndigits = self.config.get('ndigits', 6)
        self.all_xy = self.config.get('all_xy', False)
        self.xy_distribution = self.config.get('xy_distribution', False)
        self.forward_only = self.config.get('forward_only', False)
        self.backward_only = self.config.get('backward_only', False)
        self.ur_distribution = self.config.get('ur_distribution', False)
        self.mv_distribution = self.config.get("mv_distribution", False)
        self.wg_distribution = self.config.get("wg_distribution", False)
        self.param_distribution = self.config.get("param_distribution", False)
        self.mg_direction = self.config.get('mg_direction', False)
        self.cc_distribution = self.config.get("cc_distribution", {})
        self.stack_info = self.config.get('stack_info', False)
        self.monitor_mbs_grad = self.config.get('monitor_mbs_grad', False)
        self.recording_l2_features = self.config.get("recording_l2_features", False)
        self.sa_order = self.config.get("sa_order", "s,b,h,d")

        if not self.cc_distribution.get('enable', False):
            self.cc_log_only = False
        else:
            self.cc_codeline = self.cc_distribution.get('cc_codeline', [])
            self.cc_log_only = self.cc_distribution.get('cc_log_only', False)
            self.cc_logged_stack = defaultdict(set)
            self.cc_pre_hook = self.cc_distribution.get('cc_pre_hook', False)

        self.common_info()

        # 初始化AnomalyData工厂
        alert_setting = self.config.get('alert', {"rules": []})
        self.alert_rules = AnomalyScanner.load_rules(alert_setting["rules"])
        self.anomaly_data_factory = None
        if alert_setting.get('dump', False):
            self.anomaly_data_factory = AnomalyDataFactory(self.rank, self.pp_stage, self.group_mates)

        # 初始化writer, 创建输出目录
        if self.format not in FORMAT_MAPPING:
            logger.warning(f"Unsupported format: {self.format}, use default format: {MonitorConst.CSV}")
            self.format = MonitorConst.CSV

        if self.ur_distribution and self.format != 'tensorboard':
            logger.warning("can only set ur_distribution when format is 'tensorboard', cancel ur_distribution")
            self.ur_distribution = False

        writer = FORMAT_MAPPING[self.format]
        self.step_count_per_record = self.config.get('step_count_per_record', 1)

        if (self.rank in self.module_rank_list) or len(self.module_rank_list) == 0:
            self.summary_writer = writer(
                WriterInput(
                    self.tensorboard_dir,
                    self.alert_rules,
                    self.unique_id,
                    self.anomaly_data_factory,
                    self.ndigits,
                    self.step_count_per_record
                )
            )
            # 初始化anomaly detected文件目录
            if self.anomaly_data_factory:
                self.anomaly_data_writer = AnomalyDataWriter(os.path.join(self.output_base_dir, "anomaly_detected"),
                                                             self.rank)
                self.anomaly_data_writer.init_detected_json()

    def common_info(self):
        if not self.xy_distribution:
            logger.info_on_rank_0("> module input/output input_grad/output_grad is not monitored. ")
        if self.forward_only:
            logger.info_on_rank_0("> only module forward is monitored. ")
        if not self.ur_distribution:
            logger.info_on_rank_0("> update vector and ratio vector of adam is not monitored. ")
        if not self.mv_distribution:
            logger.info_on_rank_0("> momentum and variance of adam is not monitored. ")
        if not self.wg_distribution:
            logger.info_on_rank_0("> weight grad of specified module is not monitored. ")
        if not self.recording_l2_features:
            logger.info_on_rank_0("> l2 features of specified module is not monitored. ")
        if not self.mg_direction:
            logger.info_on_rank_0('> grad and momentum direction will not be compared.')
        if not self.cc_distribution.get('enable', False):
            logger.info_on_rank_0("> cc operator is not monitored.")

    # 保留原接口, 兼容msprobe1.2.2前旧版本
    def monitor_gnorm_with_ad(self, model, optimizer=None, grad_acc_steps=1, tp_group=None, dp_group=None,
                              start_iteration=0):
        if optimizer is None:
            optimizer = getattr(self, "optimizer_trans", None)  # 兼容老版本可传None的情况, 从set_wrapped_optimizer获取
            if optimizer is None:
                logger.error("monitor_gnorm_with_ad: please set_wrapped_optimizer before it or input optimizer!=None")
                return
        self.set_monitor(model, optimizer, grad_acc_steps, tp_group, dp_group, start_iteration)

    # 保留原接口, 兼容msprobe1.2.2前旧版本
    def set_wrapped_optimizer(self, optimizer):
        self.optimizer_trans = optimizer

    def set_monitor(
            self,
            model,
            optimizer,
            grad_acc_steps=1,
            tp_group=None,
            dp_group=None,
            start_iteration=0
    ):
        """External interface"""
        grad_acc_steps, start_iteration = validate_set_monitor(grad_acc_steps, start_iteration)
        global start_step
        start_step = start_iteration
        logger.info(f'grad acc steps {grad_acc_steps}')
        self.micro_batch_number = grad_acc_steps
        self.dp_group = dp_group
        self.tp_group = tp_group
        self.optimizer_mon = OptimizerMonFactory.create_optimizer_mon(optimizer)
        self.hook_step_final(optimizer)
        if not isinstance(model, list):
            model = [model]
        self.model = model
        if len(model) > 1:
            self.vpp = True
            self._smallest_rank_print('vpp enabled')
        if not self.dynamic_enable:
            self.register_hooks(optimizer)

    def register_hooks(self, optimizer):
        self._register_param_name()
        self.hook_optimizer(optimizer)
        self._patch_grad_sync()
        self.hook_modules()
        if self.cc_distribution.get('enable', False):
            self.handles['cc'] = api_register.initialize_hook(*create_hooks(context=self.cc_context, monitor=self))
            api_register.redirect_api()
        self.monitoring = True

    def adhoc_check(self, target_tensor: torch.tensor, module_name: str, tensor_name: str, rank_list, ops_list):
        rank = None
        if dist.is_initialized():
            rank = dist.get_rank()
            if (rank not in rank_list) and len(rank_list) != 0:
                return
        self.tensor_metrics.stat_insert(target_tensor, ops_list, module_name, tensor_name, rank)

    def build_tbtag_tensor_map(self, module_name, suffix, tag, tensor):
        """
        :param module_name: str of module name
        :param suffix:
        :param tag:
        :param tensor: torch.tensor or tuple/list of torch.tensor
        :return: tensor_map
        """
        tensor_map = {}
        if isinstance(tensor, torch.Tensor):
            tensor = [tensor]
        if isinstance(tensor, tuple) or isinstance(tensor, list):
            if len(tensor) == 1:
                key = get_summary_writer_tag_name(module_name + suffix, tag, self.rank)
                self.register_param_call_id("_hook_module", key)
                tensor_map[key] = tensor[0]
            else:
                for i, tensor_i in enumerate(tensor):
                    key = get_summary_writer_tag_name(module_name + f"_{i}" + suffix, tag, self.rank)
                    self.register_param_call_id("_hook_module", key)
                    tensor_map[key] = tensor_i
        return tensor_map

    def generate_param_map(self, tag, param_tensor):
        metrics = {}
        for name in self.param2name.values():
            key = get_summary_writer_tag_name(name, tag, self.rank)
            self.register_param_call_id("optimizer_pre_step_hook", key)
            if name not in param_tensor or param_tensor[name] is None:
                continue
            metrics[key] = param_tensor[name]
        return metrics

    def generate_param_metrics(self, opt_context, stage=MonitorConst.PRE_PARAM):
        if not self.param_distribution:
            return
        tag2param = {
            self.name2tag.get(name, {}).get(stage): param
            for name, param in self.name2param.items()
            if param.numel() != 0
        }
        get_metrics(self.ops, tag2param, self.eps, opt_context.param_metric)

    def generate_mv_metrics(self, opt_context):
        if not self.mv_distribution:
            return
        opt_context.exp_avg_metric = {}
        opt_context.exp_avg_sq_metric = {}
        m_tag_tensor_map = self.generate_param_map(MonitorConst.EXP_AVG, opt_context.param_exp_avg)
        v_tag_tensor_map = self.generate_param_map(MonitorConst.EXP_AVG_SQ, opt_context.param_exp_avg_sq)
        get_metrics(self.ops, m_tag_tensor_map, self.eps, opt_context.exp_avg_metric)
        get_metrics(self.ops, v_tag_tensor_map, self.eps, opt_context.exp_avg_sq_metric)

    def generate_wgrad_metrics(self, post_grad_dict):
        if not self.wg_distribution:
            return {}, {}

        get_metrics(self.ops, post_grad_dict, self.eps, self.grad_context.post)
        return self.grad_context.post, self.grad_context.pre

    def generate_xy_metrics(self):
        actv = {}
        for fwd_context in self.module_fwd_hook_context_by_module.values():
            actv.update(fwd_context.actv)

        actv_grad = self.grad_context.actv
        return actv, actv_grad

    def reload_xy(self, xy_distribution=False):
        logger.warning("reload_xy() is deprecated and will be removed in a future version. "
                       "Use DYNAMIC_MONITOR instead.")
        self.xy_distribution = xy_distribution

        for handle in self.handles['xy']:
            handle.remove()
        self.handles['xy'].clear()
        self.hook_modules()
        for _, fwd_context in self.module_fwd_hook_context_by_module.items():
            fwd_context.actv.clear()

    def write_adhoc_check(self, step):
        self.tensor_metrics.flush(self.summary_writer)

    def write_stack_info(self):
        stack_data = []
        header = ["module_name", "stack_info"]
        stack_data.append(header)
        for _, fwd_context in self.module_fwd_hook_context_by_module.items():
            stack_data.append([fwd_context.module_name, fwd_context.stack])
        filepath = os.path.join(self.tensorboard_dir, f'stack_info.csv')
        if not os.path.exists(filepath):
            data_frame = pd.DataFrame(columns=stack_data)
            write_df_to_csv(data_frame, filepath)

    def write_xy_tb(self, step):
        if not self.xy_distribution:
            return
        for _, fwd_context in self.module_fwd_hook_context_by_module.items():
            if len(fwd_context.actv) == 0:
                continue
            self.summary_writer.write_metrics(self.ops, fwd_context.actv, step, MonitorConst.ACTV)
            fwd_context.actv.clear()
        if self.grad_context.actv:
            self.summary_writer.write_metrics(self.ops, self.grad_context.actv, step, MonitorConst.ACTVGRAD)

    def write_metrics_if_not_empty(self, features, metrics, step, hook_name):
        if not features or len(features) == 0:
            return
        use_micro_step = hook_name not in ["linear_hook"]
        self.summary_writer.write_metrics(metrics, features, step, hook_name, use_micro_step=use_micro_step)
        features.clear()

    def write_features_tb(self, step):
        if not self.recording_l2_features:
            return
        for context in self.feature_hook_context_by_module.values():
            num_features = len(context.attention_feature) + len(context.linear_feature)
            if num_features == 0:
                continue
            self.write_metrics_if_not_empty(context.attention_feature, ["entropy", "softmax_max"],
                                            step, "attention_hook")
            self.write_metrics_if_not_empty(context.linear_feature, ["sr", "kernel_norm"], step, "linear_hook")

    def write_param_tb(self, opt_context):
        if not self.param_distribution:
            return
        param_metrics = {k: v for k, v in opt_context.param_metric.items() if MonitorConst.PRE_PARAM in k}
        updated_param_metrics = {k: v for k, v in opt_context.param_metric.items() if MonitorConst.POST_PARAM in k}
        self.summary_writer.write_metrics(self.ops, param_metrics, opt_context.step, MonitorConst.PRE_PARAM)
        self.summary_writer.write_metrics(self.ops, updated_param_metrics, opt_context.step, MonitorConst.POST_PARAM)

    def write_mv_tb(self, opt_context):
        if not self.mv_distribution:
            return
        self.summary_writer.write_metrics(self.ops, opt_context.exp_avg_metric,
                                          opt_context.step, MonitorConst.EXP_AVG)
        self.summary_writer.write_metrics(self.ops, opt_context.exp_avg_sq_metric,
                                          opt_context.step, MonitorConst.EXP_AVG_SQ)

    def write_grad_tb(self, step):
        if not self.wg_distribution:
            return

        self.summary_writer.write_metrics(self.ops, self.grad_context.pre, step, 'grad_unreduced',
                                          use_micro_step=self.monitor_mbs_grad)
        self.summary_writer.write_metrics(self.ops, self.grad_context.post, step, 'grad_reduced')

    def hook_optimizer(self, optimizer):
        # in DDP by default use params_have_main_grad
        def optimizer_pre_step_hook(optimizer, args, kwargs):
            context = self.optimizer_context[optimizer]

            if (self.print_struct and not all(value == {} for value in self.module_struct.values())
                    and not self.struct_printed):
                self._save_module_struct()
                if not self.cc_log_only:
                    raise Exception("exit after first monitor step when print model struct")
            if self.cc_log_only and context.step > 0:
                self._smallest_rank_print("> Used communication ops and corresponding stack")
                self._smallest_rank_print(
                    json.dumps({k: [i.split(';') for i in v] for k, v in self.cc_logged_stack.items()}))
                raise Exception("exit after first step when print cc stack")

            # skip generate metrics
            if context.step < self.start_step or (context.step - self.start_step) % self.step_interval != 0:
                return

            grad_dict = {}
            if self.wg_distribution:
                grad_dict = self.optimizer_mon.fetch_grad(self, self.param2name)

            mv_result = None
            if self.mv_distribution or self.ur_distribution or self.mg_direction:
                mv_result = self.optimizer_mon.fetch_mv(self, self.param2name)
            if mv_result:
                context.param_exp_avg = mv_result.exp_avg
                context.param_exp_avg_sq = mv_result.exp_avg_sq
                context.param_adam_update = mv_result.update
                context.param_adam_ratio = mv_result.ratio

            _, _ = self.generate_wgrad_metrics(grad_dict)
            self.generate_mv_metrics(context)
            self.generate_param_metrics(context, MonitorConst.PRE_PARAM)

            tbtag_tensor_map = {}
            if self.mg_direction:
                for param, name in self.param2name.items():
                    grad = param.main_grad if self.params_have_main_grad else param.grad
                    if grad is None:
                        logger.warning(f"grad is None: {name}, maybe something wrong happened.")
                        continue
                    if context.step == 0:
                        same_direction_ratio = torch.tensor(1.)
                    else:
                        same_direction_ratio = get_sign_matches(grad, context.param_exp_avg[name])
                    context.param_mg_direction[name] = same_direction_ratio
                tbtag_tensor_map.update(self.generate_param_map('mg_direction', context.param_mg_direction))

            metric_dict = {}
            get_metrics(self.ops, tbtag_tensor_map, self.eps, metric_dict)
            for cc in self.cc_context.values():
                cc.aggregate()
                metric_dict.update(cc.data)
                cc.reset()

            if not metric_dict:
                return
            context.metric_dict = metric_dict
            return

        def optimizer_post_step_hook(optimizer, args, kwargs):
            context = self.optimizer_context[optimizer]
            self.generate_param_metrics(context, MonitorConst.POST_PARAM)

        if self.optimizer_hooked:
            return

        self.pre_step_hooks.append(optimizer_pre_step_hook)
        self.post_step_hooks.append(optimizer_post_step_hook)

        self.optimizer_hooked = True
        return

    def dynamic_monitor(self, optimizer):
        """
        If dynamic monitor enabled and config.json updated,
        remove hooks and register new hooks according to new configuration.
        """
        context = self.optimizer_context[optimizer]
        if not self.dynamic_enable:
            return
        try:
            # 如果文件时间戳没变, 可以不读取节省时间
            config_timestamp = os.path.getmtime(self.config_file_path)
            if config_timestamp == self.config_timestamp:
                return
            # 更新config文件最新修改时间戳
            self.config_timestamp = config_timestamp
            config = load_json(self.config_file_path)
        except Exception as e:
            logger.error(f"get config.json wrong because {e}, not updated, please check!!!")
            return

        if config.get("dynamic_on", False):
            try:
                validate_config(config)
                self.config = config
                self.set_config()
                self.start_step = context.step  # 动态启停时不受原start_step影响，永远从下一步开始
                logger.warning(f"config is updated at step{context.step - 1}, "
                               f"will start new hook at step{context.step}.")
            except Exception as e:
                logger.error(f"set config wrong because {e}, not updated, please check!!!")
                return

            self._remove_all_hooks(optimizer)
            self.register_hooks(optimizer)

    def hook_step_final(self, optimizer):
        def step_final_hook(optimizer, args, kwargs):
            context = self.optimizer_context[optimizer]
            rank = dist.get_rank() if dist.is_initialized() else None
            # 静态在第0步就可以保存, 动态在第0步不可以, 因为动态设计的就是重置后下一步开启, 第0步的self.monitoring还是False
            if self.monitoring:
                module_rank_valid = not self.module_rank_list or (
                        dist.is_initialized() and dist.get_rank() in self.module_rank_list)
                step_condition = (context.step >= self.start_step and (
                        context.step - self.start_step) % self.step_interval == 0)
                if module_rank_valid and step_condition:
                    self.has_collect_times += 1

                    if self.anomaly_data_factory:
                        self.anomaly_data_factory.set_call_id(self.param_name_call_id)
                    self.write_xy_tb(context.step)
                    self.write_features_tb(context.step)
                    self.write_grad_tb(context.step)
                    self.write_mv_tb(context)
                    self.write_param_tb(context)
                    self.write_adhoc_check(context.step)
                    if self.stack_info:
                        self.write_stack_info()
                        self.stack_info = False
                        for handle in self.handles["stack"]:
                            handle.remove()
                        self.handles["stack"].clear()

                    if self.ur_distribution:
                        for param_name, _ in context.param_adam_update.items():
                            self.update_heatmap_visualizer[param_name].visualize(
                                get_summary_writer_tag_name(param_name, 'adam_update', rank), context.step,
                                self.summary_writer)
                        for param_name, _ in context.param_adam_ratio.items():
                            self.ratio_heatmap_visualizer[param_name].visualize(
                                get_summary_writer_tag_name(param_name, 'adam_ratio', rank), context.step,
                                self.summary_writer)

                    if context.metric_dict:
                        self.summary_writer.write_metrics(self.ops, context.metric_dict, context.step, 'other')
                    context.metric_dict.clear()

                    if self.anomaly_data_factory:
                        self.anomaly_data_writer.write_detected_json(self.summary_writer.get_anomalies())
                    self.summary_writer.clear_anomalies()

                    if self.format == MonitorConst.TENSORBOARD:
                        chmod_tensorboard_dir(self.tensorboard_dir)
                    self.call_id = 0
                    self.param_name_call_id.clear()

                    if self.has_collect_times >= self.collect_times:
                        self._remove_all_hooks_final(optimizer)

            context.step += 1
            self.dynamic_monitor(optimizer)

        def patch_step(func, optimizer):
            def wrapper(*args, **kwargs):
                for hook in self.pre_step_hooks:
                    hook(optimizer, args, kwargs)
                out = func(*args, **kwargs)
                for hook in self.post_step_hooks:
                    hook(optimizer, args, kwargs)
                step_final_hook(optimizer, args, kwargs)
                return out

            return wrapper

        optimizer.__class__.step = patch_step(optimizer.__class__.step, optimizer)
        return

    def hook_modules(self):
        if self.module_rank_list and (self.rank not in self.module_rank_list):
            return

        targets = self.config['targets']
        module_in_all_stage = [key for key in targets.keys() if MonitorConst.NAME_SEP not in key]
        for key in module_in_all_stage:
            struct = targets.pop(key)
            targets.update({f'{vpp_stage}{MonitorConst.NAME_SEP}{key}': struct for vpp_stage in range(len(self.model))})

        hooked_count = 0
        for vpp_stage, model_chunk in enumerate(self.model):
            vpp_stage = f'{vpp_stage}{MonitorConst.NAME_SEP}'
            targets = [x for x, _ in model_chunk.named_modules()] if self.print_struct else self.config[
                'targets'].keys()
            l2_target_names = self.config.get('l2_targets', '')
            hooked_count += self._hook_module(targets, l2_target_names, model_chunk, vpp_stage)

        logger.info_on_rank_0(f"> {hooked_count} modules are monitored.")

        @recursion_depth_decorator('msprobe.pytorch.monitor.clone_if_tensor')
        def clone_if_tensor(args):
            if isinstance(args, tuple):
                return tuple([clone_if_tensor(arg) for arg in args])
            elif isinstance(args, torch.Tensor):
                return args.clone()
            else:
                return args

        @torch.no_grad
        def wrap_hook_setup(setup):
            def wrapped_setup(*args, **kwargs):
                args = setup(*args, **kwargs)
                args = clone_if_tensor(args)
                return args

            return wrapped_setup

        BackwardHook.setup_input_hook = wrap_hook_setup(BackwardHook.setup_input_hook)
        BackwardHook.setup_output_hook = wrap_hook_setup(BackwardHook.setup_output_hook)
        return

    def register_param_call_id(self, hook_name: str, key: str):
        """
        :param hook_name:
        :param key: str, '0:relu_0/output_grad'
        :return:
        """
        logger.debug(f"{hook_name} {key}: {self.call_id}")
        self.param_name_call_id[key] = self.call_id
        self.call_id += 1

    def _remove_all_hooks(self, optimizer):
        # 清空hook handle
        for handle in self.handles['xy']:
            handle.remove()
        self.handles['xy'].clear()
        for handle in self.handles['L2_features']:
            handle.remove()
        self.handles['L2_features'].clear()
        # 清空对应context缓存
        for _, fwd_context in self.module_fwd_hook_context_by_module.items():
            fwd_context.reset()
        for _, bwd_context in self.module_bwd_hook_context_by_module.items():
            bwd_context.reset()
        self.grad_context.reset()  # 权重梯度和激活值梯度都在这

        # 还原梯度patch
        self.optimizer_mon.restore_grad_sync(self)
        if self.fsdp_post_backward_hook:  # fsdp
            torch.distributed.fsdp._runtime_utils._post_backward_hook = self.fsdp_post_backward_hook
            logger.info("remove patch_post_backward_hook in fsdp.")
        if self.fsdp2_foreach_reduce:  # fsdp2
            torch.distributed.fsdp._fully_shard._fsdp_collectives.foreach_reduce = self.fsdp2_foreach_reduce
            importlib.reload(torch.distributed.fsdp._fully_shard._fsdp_param_group)
            logger.info("remove patch_foreach_reduce_hook in fsdp2.")
        else:  # not megatron and not fsdp
            for handle in self.handles['wgrads']:
                handle.remove()
            self.handles['wgrads'].clear()
            self.weight_hooked = False

        if self.optimizer_hooked:
            self.pre_step_hooks.clear()
            self.post_step_hooks.clear()

        for _, context in self.optimizer_context.items():
            context.reset()
        self.optimizer_hooked = False

        for handle in self.handles['cc']:
            handle.remove()
        self.handles['cc'].clear()
        api_register.restore_api()
        for _, context in self.cc_context.items():
            context.reset()

        # 清空节点缓存
        self.param2name.clear()
        self.name2indices.clear()
        self.name2param.clear()
        self.duplicate_param.clear()
        self.name2tag.clear()
        self.module_struct.clear()
        self.grad_accs.clear()

        # 关闭采集状态
        self.monitoring = False

    def _remove_all_hooks_final(self, optimizer):
        if self.dynamic_enable:
            # 结束后自动重置dynamic_on为False等待用户手动开启
            try:
                config = load_json(self.config_file_path)
                config['dynamic_on'] = False
                save_json(self.config_file_path, config, indent=2)
                config_timestamp = os.path.getmtime(self.config_file_path)
                self.config_timestamp = config_timestamp
                logger.info(
                    "Finish monitor, set config'dynamic_on=False, will restart by set it to True and update config")
            except Exception as e:
                logger.warning(f"Finish monitor, set config'dynamic_on=False fail because {e}, please check!!!")
        logger.info("Finish monitor")
        self._remove_all_hooks(optimizer)

    def _smallest_rank_print(self, msg):
        if dist.is_initialized():
            if self.module_rank_list:
                if dist.get_rank() == min(self.module_rank_list):
                    logger.info(msg)
            else:
                if dist.get_rank() == 0:
                    logger.info(msg)
        else:
            logger.info(msg)

    def _save_module_struct(self):
        output_dir = os.path.join(get_output_base_dir(), 'module_struct', f'rank{self.rank}')
        create_directory(output_dir)
        module_struct_file = os.path.realpath(os.path.join(output_dir, 'module_struct.json'))
        save_json(module_struct_file, self.module_struct, indent=2)
        logger.info(f"> save module struct to {module_struct_file}")
        self.struct_printed = True

    def _is_target_param(self, param_name, param, prefix):
        name = prefix + param_name
        squash_name = prefix + squash_param_name(param_name, self.squash_name)
        for target in self.config['targets'].keys():
            if param_name.startswith(target) or squash_name.startswith(target) or name.startswith(target):
                return True

        return False

    def _register_chunk(self, model_chunk, prefix):
        if isinstance(model_chunk, FSDP):
            if not model_chunk._use_orig_params:
                raise ValueError("Only Support fsdp1 with use_orig_params=True")
            self.fsdp_wrapped_module = True
        for (param_name, param) in model_chunk.named_parameters():
            if not param.requires_grad:
                continue
            if not self.fsdp2_wrapped_module and param.__class__.__name__ == "DTensor":
                self.fsdp2_wrapped_module = True
            if self.fsdp_wrapped_module:  # FSDP1需要记录完整的不被target限制的flat权重名，以供后续对flat解包
                self.fsdp_param_name_map[id(param)] = param_name

            if self._is_target_param(param_name, param, prefix):
                name = prefix + squash_param_name(param_name, self.squash_name)
                if name in self.param2name.values():
                    name = prefix + param_name
                self.param2name[param] = name
                self.name2param[name] = param
                self.origin2squash[param_name] = name

                if self.tp_group and not param_is_not_tensor_parallel_duplicate(param, self.tp_group):
                    self.duplicate_param[name] = True
                if self.dp_group and param_is_data_parallel_duplicate(self.dp_group):
                    self.duplicate_param[name] = True

                keywords = [
                    MonitorConst.PRE_GRAD,
                    MonitorConst.POST_GRAD,
                    MonitorConst.PRE_PARAM,
                    MonitorConst.POST_PARAM
                ]
                self.name2tag[name] = {
                    k: get_summary_writer_tag_name(name, k, self.rank)
                    for k in keywords
                }

    def _register_param_name(self):
        for vpp_stage, model_chunk in enumerate(self.model):
            prefix = f'{vpp_stage}{MonitorConst.NAME_SEP}'
            self._register_chunk(model_chunk, prefix)

    def _is_target_module(self, module_name, targets, vpp_stage):
        if self.all_xy or self.print_struct:
            return vpp_stage + squash_param_name(module_name, self.squash_name)
        for pattern in [
            vpp_stage + squash_param_name(module_name, self.squash_name),
            vpp_stage + module_name,
        ]:
            if pattern in targets:
                return pattern
        return ""

    def _is_recording_module(self, module_name, l2_targets, vpp_stage, hook_name):

        if len(l2_targets) > 0:
            for pattern in [
                vpp_stage + squash_param_name(module_name, self.squash_name),
                vpp_stage + module_name,
            ]:
                if pattern in l2_targets:
                    return pattern
        elif hook_name in ["linear_hook"]:
            return vpp_stage + squash_param_name(module_name, self.squash_name)
        return ""

    def _hook_module(self, target_names, l2_target_names, module: torch.nn.Module, vpp_stage=''):
        if '_modules' not in module.__dict__:
            # nothing to hook
            return 0

        def fwd_hook_fun(module, args, kwargs, module_output, name):
            if not module.training or is_recomputation():
                # 1 only monitor training stage.
                # 2 when open recompute, skip recomputed forward stage.
                return

            module_input = [tensor for tensor in args if torch.is_tensor(tensor)]
            if kwargs:
                kwargs_tensors = [tensor for tensor in kwargs.values() if torch.is_tensor(tensor)]
                module_input.extend(kwargs_tensors)

            if module not in self.module_fwd_hook_context_by_module:
                self.module_fwd_hook_context_by_module[module] = ModuleHookContext(name)
            context: ModuleHookContext = self.module_fwd_hook_context_by_module[module]
            if not context.struct:
                context.struct = {
                    Const.INPUT: get_param_struct(module_input),
                    Const.OUTPUT: get_param_struct(module_output)
                }

            if self.print_struct:
                self.module_struct[context.module_name].update(context.struct)
                return

            tbtag_tensor_map = {}
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.INPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTV, module_input))
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.OUTPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTV, module_output))

            get_metrics(self.ops, tbtag_tensor_map, self.eps, context.actv)
            context.micro_step += 1
            if context.micro_step == self.micro_batch_number:
                context.micro_step = 0
            return

        def bwd_hook_fun(module, input_grad, output_grad):
            context: ModuleHookContext = self.module_bwd_hook_context_by_module[module]
            if not context.struct:
                context.struct = {
                    MonitorConst.INPUT_GRAD: get_param_struct(input_grad),
                    MonitorConst.OUTPUT_GRAD: get_param_struct(output_grad)
                }
            if self.print_struct:
                self.module_struct[context.module_name].update(context.struct)
                return

            tbtag_tensor_map = {}
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.INPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTVGRAD, input_grad))

            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.OUTPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTVGRAD, output_grad))

            if context.micro_step == 0 and context.actvgrad:
                logger.warning(f"actvgrad context of {context.module_name} is not empty when first micro_step, "
                               f"maybe something wrong happened. Now clear it.")
                context.actvgrad.clear()

            get_metrics(self.ops, tbtag_tensor_map, self.eps, self.grad_context.actv)

            context.micro_step += 1
            if context.micro_step == self.micro_batch_number:
                context.micro_step = 0
            return

        def extract_attention_feature_hook(module, module_input, module_output, name):
            if is_recomputation() or not module.training:
                return

            if module not in self.feature_hook_context_by_module:
                self.feature_hook_context_by_module[module] = FeatureHookContext(name)
            context: FeatureHookContext = self.feature_hook_context_by_module[module]
            tbtag_tensor_map = {}
            if len(module_input) < 2:
                logger.warning(
                    f"Length of module_input in attention hook ({name}) is {len(module_input)}, "
                    "expected >= 2. Skipping feature extraction for this module."
                )
                return
            q_h = module_input[0]
            k_h = module_input[1]
            qkt = cal_qkt(q_h, k_h, order=self.sa_order)
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(f'{context.module_name}.attention',
                                            f'{MonitorConst.NAME_SEP}{context.micro_step}', 'qkt', qkt)
            )
            get_entropy_metric(tbtag_tensor_map, context.attention_feature)

            context.micro_step += 1
            if context.micro_step == self.micro_batch_number:
                context.micro_step = 0
                context.step += 1
            return

        def extract_linear_sr_hook(module, module_input, module_output, name):
            if is_recomputation() or not module.training:
                return
            weight_name = self.get_linear_hook_target(module)
            if weight_name == '':
                return

            if module not in self.feature_hook_context_by_module:
                self.feature_hook_context_by_module[module] = FeatureHookContext(name)
            context: FeatureHookContext = self.feature_hook_context_by_module[module]

            if context.micro_step == (self.micro_batch_number - 1):
                tbtag_tensor_map = {}
                value = getattr(module, weight_name).data
                tbtag_tensor_map.update(
                    self.build_tbtag_tensor_map(f'{context.module_name}.linear',
                                                '', 'sr', value)
                )
                get_sr_metric(tbtag_tensor_map, context.linear_feature)

            context.micro_step += 1
            if context.micro_step == self.micro_batch_number:
                context.micro_step = 0
                context.step += 1
            return

        def stack_hook(module, args, kwargs, module_output, name):
            if module not in self.module_fwd_hook_context_by_module:
                self.module_fwd_hook_context_by_module[module] = ModuleHookContext(name)
            context: ModuleHookContext = self.module_fwd_hook_context_by_module[module]
            context.stack = analyze_api_call_stack(name)
            return

        if self.backward_only and self.forward_only:
            logger.warning('not enable backward_only and forward_only simultaneously')

        hooked_count = 0
        for module_name, submodule in module.named_modules():
            if self.stack_info:
                name = vpp_stage + squash_param_name(module_name, self.squash_name)
                handle = submodule.register_forward_hook(partial(stack_hook, name=name), with_kwargs=True)
                self.handles['stack'].append(handle)
            name = self._is_target_module(module_name, target_names, vpp_stage)
            if not name:
                continue
            if submodule.__class__.__name__ == "FullyShardedDataParallel":
                continue
            if self.xy_distribution or self.print_struct:
                if not self.backward_only:
                    handle = submodule.register_forward_hook(partial(fwd_hook_fun, name=name), with_kwargs=True)
                    self.handles['xy'].append(handle)
                if not self.forward_only and not self.has_register_backward_hook(name, submodule):
                    handle = submodule.register_full_backward_hook(bwd_hook_fun)
                    self.handles['xy'].append(handle)
                    self.module_bwd_hook_context_by_module[submodule] = ModuleHookContext(name)
                logger.info_on_rank_0(f"> {name} is monitored successfully")
                hooked_count += 1
        if not self.print_struct and self.recording_l2_features:
            for module_name, submodule in module.named_modules():
                func_map = {
                    "attention_hook": extract_attention_feature_hook,
                    "linear_hook": extract_linear_sr_hook,
                }
                for hook_name in func_map.keys():
                    if hook_name not in l2_target_names:
                        continue
                    temp_names = l2_target_names[hook_name]
                    name = self._is_recording_module(module_name, temp_names, vpp_stage, hook_name)
                    if name:
                        handle = submodule.register_forward_hook(partial(func_map[hook_name], name=name))
                        print_feature_name = hook_name.split('_')[0]
                        logger.info_on_rank_0(
                            f'> {print_feature_name} features of {name} is monitored successfully')
                        self.handles["L2_features"].append(handle)
                        hooked_count += 1
                continue

        return hooked_count

    def _patch_grad_sync(self):
        if not self.wg_distribution:
            return

        if self.fsdp_wrapped_module:
            # patch fsdp _runtime_utils._post_backward_hook
            self._patch_fsdp_post_backward_hook()
            return

        if self.fsdp2_wrapped_module:
            # patch fsdp2 _fully_shard._fsdp_collectives.foreach_reduce
            self._patch_fsdp2_foreach_reduce()
            return

        if self.monitor_mbs_grad:
            self._hook_weights()
            return

        self.optimizer_mon.patch_grad_sync(self)

        if self.enable_megatron or self.enable_deepspeed:
            return

        # default hook weights
        self._hook_weights()

    def _patch_fsdp_post_backward_hook(self):
        """
        FSDP runtime 需要处理整个forward和backward计算和通信的流程，通过override nn.Module的forward，定义相应的逻辑。
        对AccumulateGrad对象注册hook，可以在backward计算grad后立刻执行，在reduce_scatter操作前采集梯度累计后，通信聚合前的梯度。
        每个forward阶段，fsdp对AccumulateGrad重复注册hook方法，monitor工具内注册hook无法生效，
        因此对_post_backward_hook进行patch，在backward后，reduce_scatter前采集梯度。
        """

        def patch_post_backward_hook(_post_backward_hook):
            def wrapper(state, handle, *unused):
                if not hasattr(handle.flat_param, 'micro_step'):
                    setattr(handle.flat_param, 'micro_step', 0)

                handle_not_in_any_model = all(
                    model._all_handles is None or handle not in model._all_handles
                    for model in self.model
                )
                if handle_not_in_any_model:
                    logger.debug(f"Skip handle not along to the monitor model list, handle id: {id(handle)}.")
                elif self.monitor_mbs_grad or (handle.flat_param.micro_step + 1 == self.micro_batch_number):
                    grad_dict = {}
                    offsets = handle._get_flat_param_offsets()
                    shapes = handle.flat_param._shapes
                    params = handle.flat_param._params
                    for param, (start, end), local_shape in zip(params, offsets, shapes):
                        param_name = self.fsdp_param_name_map.get(id(param), "")
                        tag = self.name2tag.get(self.origin2squash.get(param_name, ""), {}).get(MonitorConst.PRE_GRAD)
                        if tag is None:
                            continue
                        grad = handle.flat_param.grad[start:end + 1].reshape(local_shape)
                        if self.monitor_mbs_grad:
                            tag = tag.replace('/', f'{MonitorConst.NAME_SEP}{handle.flat_param.micro_step}/', 1)
                        grad_dict[tag] = grad
                        self.register_param_call_id("_post_backward_hook", tag)
                    get_metrics(self.ops, grad_dict, self.eps, self.grad_context.pre)
                handle.flat_param.micro_step = (handle.flat_param.micro_step + 1) % self.micro_batch_number
                out = _post_backward_hook(state, handle, *unused)
                return out

            return wrapper

        logger.info("Patch fsdp _post_backward_hook, collect pre_grad metrics.")
        self.fsdp_post_backward_hook = torch.distributed.fsdp._runtime_utils._post_backward_hook
        torch.distributed.fsdp._runtime_utils._post_backward_hook = \
            patch_post_backward_hook(torch.distributed.fsdp._runtime_utils._post_backward_hook)

    def _patch_fsdp2_foreach_reduce(self):
        def patch_foreach_reduce(foreach_reduce):
            def wrapper(fsdp_params, unsharded_grads, *unused):
                if not hasattr(fsdp_params[0], 'micro_step'):
                    setattr(fsdp_params[0], 'micro_step', 0)

                if self.monitor_mbs_grad or (fsdp_params[0].micro_step + 1 == self.micro_batch_number):
                    grad_dict = {}
                    for param, grad in zip(fsdp_params, unsharded_grads):
                        tag = self.name2tag.get(self.origin2squash.get(param._param_fqn, ""), {}).get(MonitorConst.PRE_GRAD)
                        if tag is None:
                            continue
                        if self.monitor_mbs_grad:
                            tag = tag.replace('/', f'{MonitorConst.NAME_SEP}{fsdp_params[0].micro_step}/', 1)
                        grad_dict[tag] = grad
                        self.register_param_call_id("foreach_reduce", tag)
                    get_metrics(self.ops, grad_dict, self.eps, self.grad_context.pre)
                fsdp_params[0].micro_step = (fsdp_params[0].micro_step + 1) % self.micro_batch_number
                out = foreach_reduce(fsdp_params, unsharded_grads, *unused)
                return out

            return wrapper

        logger.info("Patch fsdp2 foreach_reduce, collect pre_grad metrics.")
        import torch.distributed.fsdp._fully_shard._fsdp_param_group as _fsdp_param_group
        import torch.distributed.fsdp._fully_shard._fsdp_collectives as _fsdp_collectives
        self.fsdp2_foreach_reduce = _fsdp_collectives.foreach_reduce
        _fsdp_collectives.foreach_reduce = patch_foreach_reduce(_fsdp_collectives.foreach_reduce)
        importlib.reload(_fsdp_param_group)  # 关键操作，不然会因为torch一开始就import foreach_reduce导致patch失效

    def _hook_weights(self):
        """
        遍历参数的梯度生成函数（grad_acc），并挂载hook，以便在该参数所有梯度计算后，采集通信聚合前梯度数据。
        """

        @torch.no_grad
        def param_hook(*args, param, name):
            key = name
            if self.monitor_mbs_grad:
                key += f'{MonitorConst.NAME_SEP}{param.micro_step}'

            key = get_summary_writer_tag_name(key, 'acc_grad', self.rank)
            self.register_param_call_id("param_hook", key)
            if self.monitor_mbs_grad or (param.micro_step + 1 == self.micro_batch_number):
                if self.params_have_main_grad:
                    grad = param.main_grad
                else:
                    grad = param.grad
                get_metrics(self.ops, {key: grad.clone()}, self.eps, self.grad_context.pre)
            param.micro_step = (param.micro_step + 1) % self.micro_batch_number

        logger.info("hooking weight grads.")
        for param, name in self.param2name.items():
            setattr(param, 'micro_step', 0)
            param_tmp = param.expand_as(param)
            grad_acc = param_tmp.grad_fn.next_functions[0][0]
            handle = grad_acc.register_hook(
                partial(param_hook, param=param, name=name))
            self.grad_accs.append(grad_acc)
            self.handles['wgrads'].append(handle)
        self.weight_hooked = True
